from collections import defaultdict
from typing import Iterator, List, Tuple, TypeVar, Union

from clarifai_grpc.grpc.api import resources_pb2

from clarifai.constants.dataset import DATASET_UPLOAD_TASKS
from clarifai.datasets.upload.features import (MultiModalFeatures, TextFeatures,
                                               VisualClassificationFeatures,
                                               VisualDetectionFeatures, VisualSegmentationFeatures)

OutputFeaturesType = TypeVar(
    'OutputFeaturesType',
    bound=Union[TextFeatures, VisualClassificationFeatures, VisualDetectionFeatures,
                VisualSegmentationFeatures, MultiModalFeatures])


class ClarifaiDataset:
  """Clarifai datasets base class."""

  def __init__(self, data_generator: 'ClarifaiDataLoader', dataset_id: str,
               max_workers: int = 4) -> None:
    self.data_generator = data_generator
    self.dataset_id = dataset_id
    self.max_workers = max_workers
    self.all_input_ids = {}
    self._all_input_protos = {}
    self._all_annotation_protos = defaultdict(list)

  def __len__(self) -> int:
    """Get size of all input protos"""
    return len(self.data_generator)

  def _to_list(self, input_protos: Iterator) -> List:
    """Parse protos iterator to list."""
    return list(input_protos)

  def _extract_protos(self) -> None:
    """Create input image protos for each data generator item."""
    raise NotImplementedError()

  def get_protos(self, input_ids: List[int]
                ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
    """Get input and annotation protos based on input_ids.
    Args:
      input_ids: List of input IDs to retrieve the protos for.
    Returns:
      Input and Annotation proto iterators for the specified input IDs.
    """
    input_protos, annotation_protos = self._extract_protos(input_ids)

    return input_protos, annotation_protos


class ClarifaiDataLoader:
  """Clarifai data loader base class."""

  def __init__(self) -> None:
    pass

  @property
  def task(self):
    raise NotImplementedError("Task should be one of {}".format(DATASET_UPLOAD_TASKS))

  def load_data(self) -> None:
    raise NotImplementedError()

  def __len__(self) -> int:
    raise NotImplementedError()

  def __getitem__(self, index: int) -> OutputFeaturesType:
    raise NotImplementedError()



TRAINABLE_MODEL_TYPES = [
    'visual-classifier', 'visual-detector', 'visual-segmenter', 'visual-embedder', 'clusterer',
    'text-classifier', 'embedding-classifier', 'text-to-text'
]
MAX_MODEL_PREDICT_INPUTS = 128
MODEL_EXPORT_TIMEOUT = 1800
MIN_RANGE_SIZE = 4194304  # 4MB
MAX_RANGE_SIZE = 314572800  # 300MB
MIN_CHUNK_SIZE = 131072  # 128KB
MAX_CHUNK_SIZE = 10485760  # 10MB
RANGE_SIZE = 31457280  # 30MB
CHUNK_SIZE = 1048576  # 1MB



import os
from enum import Enum
from typing import List, Tuple, Union

from clarifai.client.dataset import Dataset
from clarifai.client.model import Model

from .helpers import (MACRO_AVG, EvalType, _BaseEvalResultHandler, get_eval_type,
                      make_handler_by_type)

try:
  import seaborn as sns
except ImportError:
  raise ImportError("Can not import seaborn. Please run `pip install seaborn` to install it")

try:
  import matplotlib.pyplot as plt
except ImportError:
  raise ImportError("Can not import matplotlib. Please run `pip install matplotlib` to install it")

try:
  import pandas as pd
except ImportError:
  raise ImportError("Can not import pandas. Please run `pip install pandas` to install it")

try:
  from loguru import logger
except ImportError:
  from ..logging import logger

__all__ = ['EvalResultCompare']


class CompareMode(Enum):
  MANY_MODELS_TO_ONE_DATA = 0
  ONE_MODEL_TO_MANY_DATA = 1


class EvalResultCompare:
  """Compare evaluation result of models against datasets.
  Note: The module will pick latest result on the datasets.
  and models must be same model type

  Args:
  ---
    models (Union[List[Model], List[str]]): List of Model or urls of models.
    datasets (Union[Dataset, List[Dataset], str, List[str]]): A single or List of Url or Dataset
    attempt_evaluate (bool): Evaluate when model is not evaluated with the datasets.
    auth_kwargs (dict): Additional auth keyword arguments to be passed to the Dataset and Model if using url(s)
  """

  def __init__(self,
               models: Union[List[Model], List[str]],
               datasets: Union[Dataset, List[Dataset], str, List[str]],
               attempt_evaluate: bool = False,
               auth_kwargs: dict = {}):
    assert isinstance(models, list), ValueError("Expected list")

    if len(models) > 1:
      self.mode = CompareMode.MANY_MODELS_TO_ONE_DATA
      self.comparator = "Model"
      assert isinstance(datasets, Dataset) or (
          isinstance(datasets, list) and len(datasets) == 1
      ), f"When comparing multiple models, must provide only one `datasets`. However got {datasets}"
    else:
      self.mode = CompareMode.ONE_MODEL_TO_MANY_DATA
      self.comparator = "Dataset"

    # validate models
    if all(map(lambda x: isinstance(x, str), models)):
      models = [Model(each, **auth_kwargs) for each in models]
    elif not all(map(lambda x: isinstance(x, Model), models)):
      raise ValueError(
          f"Expected all models are list of string or list of Model, got {[type(each) for each in models]}"
      )
    # validate datasets
    if not isinstance(datasets, list):
      datasets = [
          datasets,
      ]
    if all(map(lambda x: isinstance(x, str), datasets)):
      datasets = [Dataset(each, **auth_kwargs) for each in datasets]
    elif not all(map(lambda x: isinstance(x, Dataset), datasets)):
      raise ValueError(
          f"Expected datasets must be str, list of string or Dataset, list of Dataset, got {[type(each) for each in datasets]}"
      )
    # Validate models vs datasets together
    self._eval_handlers: List[_BaseEvalResultHandler] = []
    self.model_type = None
    logger.info("Initializing models...")
    for model in models:
      model.load_info()
      model_type = model.model_info.model_type_id
      if not self.model_type:
        self.model_type = model_type
      else:
        assert self.model_type == model_type, f"Can not compare when model types are different, {self.model_type} != {model_type}"
      m = make_handler_by_type(model_type)(model=model)
      logger.info(f"* {m.get_model_name(pretify=True)}")
      m.find_eval_id(datasets=datasets, attempt_evaluate=attempt_evaluate)
      self._eval_handlers.append(m)

  @property
  def eval_handlers(self):
    return self._eval_handlers

  def _loop_eval_handlers(self, func_name: str, **kwargs) -> Tuple[list, list]:
    """ Run methods of `eval_handlers[...].model`

    Args:
      func_name (str): method name, see `_BaseEvalResultHandler` child classes
      kwargs: keyword arguments of the method

    Return:
      tuple:
        - list of outputs
        - list of comparator names

    """
    outs = []
    comparators = []
    logger.info(f'Running `{func_name}`')
    for _, each in enumerate(self.eval_handlers):
      for ds_index, _ in enumerate(each.eval_data):
        func = eval(f'each.{func_name}')
        out = func(index=ds_index, **kwargs)

        if self.mode == CompareMode.MANY_MODELS_TO_ONE_DATA:
          name = each.get_model_name(pretify=True)
        else:
          name = each.get_dataset_name_by_index(ds_index, pretify=True)
        if out is None:
          logger.warning(f'{self.comparator}:{name} does not have valid data for `{func_name}`')
          continue
        comparators.append(name)
        outs.append(out)

    if self.mode == CompareMode.MANY_MODELS_TO_ONE_DATA:
      apps = set([comp.split('/')[0] for comp in comparators])
      if len(apps) == 1:
        comparators = ['/'.join(comp.split('/')[1:]) for comp in comparators]

    if not outs:
      logger.warning(f'Model type {self.model_type} does not support `{func_name}`')

    return outs, comparators

  def detailed_summary(self,
                       confidence_threshold: float = .5,
                       iou_threshold: float = .5,
                       area: str = "all",
                       bypass_const=False) -> Union[Tuple[pd.DataFrame, pd.DataFrame], None]:
    """
    Retrieve and compute popular metrics of model.

    Args:
      confidence_threshold (float): confidence threshold, applicable for classification and detection. Default is 0.5
      iou_threshold (float): iou threshold, support in range(0.5, 1., step=0.1) applicable for detection
      area (float): size of area, support {all, small, medium}, applicable for detection

    Return:
      None or tuple of dataframe: df summary per concept and total concepts

    """
    df = []
    total = []
    # loop over all eval_handlers/dataset and call its method
    outs, comparators = self._loop_eval_handlers(
        'detailed_summary',
        confidence_threshold=confidence_threshold,
        iou_threshold=iou_threshold,
        area=area,
        bypass_const=bypass_const)
    for indx, out in enumerate(outs):
      _df, _total = out
      _df[self.comparator] = [comparators[indx] for _ in range(len(_df))]
      _total['Concept'].replace(
          to_replace=['Total'], value=f'{self.comparator}:{comparators[indx]}', inplace=True)
      _total.rename({'Concept': 'Total Concept'}, axis=1, inplace=True)
      df.append(_df)
      total.append(_total)

    if df:
      df = pd.concat(df, axis=0)
      total = pd.concat(total, axis=0)
      return df, total
    else:
      return None

  def confusion_matrix(self, show=True, save_path: str = None,
                       cm_kwargs: dict = {}) -> Union[pd.DataFrame, None]:
    """Return dataframe of confusion matrix
    Args:
        show (bool, optional): Show the chart. Defaults to True.
        save_path (str): path to save rendered chart.
        cm_kwargs (dict): keyword args of `eval_handler[...].model.cm_kwargs` method.
    Returns:
        None or pd.Dataframe, If models don't have confusion matrix, return None
    """
    outs, comparators = self._loop_eval_handlers("confusion_matrix", **cm_kwargs)
    all_dfs = []
    for _, (df, anchor) in enumerate(zip(outs, comparators)):
      df[self.comparator] = [anchor for _ in range(len(df))]
      all_dfs.append(df)

    if all_dfs:
      all_dfs = pd.concat(all_dfs, axis=0)
      if save_path or show:

        def _facet_heatmap(data, **kws):
          data = data.dropna(axis=1)
          data = data.drop(self.comparator, axis=1)
          concepts = data.columns
          colnames = pd.MultiIndex.from_arrays([concepts], names=['Predicted'])
          data.columns = colnames
          ax = sns.heatmap(data, cmap='Blues', annot=True, annot_kws={"fontsize": 8}, **kws)
          ax.set_xticklabels(ax.get_xticklabels(), rotation=45, fontsize=6)
          ax.set_yticklabels(ax.get_yticklabels(), fontsize=6, rotation=0)

        temp = all_dfs.copy()
        temp.columns = ["_".join(pair) for pair in temp.columns]
        with sns.plotting_context(font_scale=5.5):
          g = sns.FacetGrid(
              temp,
              col=self.comparator,
              col_wrap=3,
              aspect=1,
              height=3,
              sharex=False,
              sharey=False,
          )
          cbar_ax = g.figure.add_axes([.92, .3, .02, .4])
          g = g.map_dataframe(
              _facet_heatmap, cbar_ax=cbar_ax, vmin=0, vmax=1, cbar=True, square=True)
          g.set_titles(col_template=str(self.comparator) + ':{col_name}', fontsize=5)
          if show:
            plt.show()
          if save_path:
            g.savefig(save_path)

    return all_dfs if isinstance(all_dfs, pd.DataFrame) else None

  @staticmethod
  def _set_default_kwargs(kwargs: dict, var_name: str, value):
    if var_name not in kwargs:
      kwargs.update({var_name: value})
    return kwargs

  @staticmethod
  def _setup_default_lineplot(df: pd.DataFrame, kwargs: dict):
    hue_order = df["concept"].unique().tolist()
    hue_order.remove(MACRO_AVG)
    hue_order.insert(0, MACRO_AVG)
    EvalResultCompare._set_default_kwargs(kwargs, "hue_order", hue_order)

    sizes = {}
    for each in hue_order:
      s = 1.5
      if each == MACRO_AVG:
        s = 4.
      sizes.update({each: s})
    EvalResultCompare._set_default_kwargs(kwargs, "sizes", sizes)
    EvalResultCompare._set_default_kwargs(kwargs, "size", "concept")

    EvalResultCompare._set_default_kwargs(kwargs, "errorbar", None)
    EvalResultCompare._set_default_kwargs(kwargs, "height", 5)

    return kwargs

  def roc_curve_plot(self,
                     show=True,
                     save_path: str = None,
                     roc_curve_kwargs: dict = {},
                     relplot_kwargs: dict = {}) -> Union[pd.DataFrame, None]:
    """Return dataframe of ROC curve
    Args:
        show (bool, optional): Show the chart. Defaults to True.
        save_path (str): path to save rendered chart.
        pr_curve_kwargs (dict): keyword args of `eval_handler[...].model.roc_curve` method.
        relplot_kwargs (dict): keyword args of `sns.relplot` except {data,x,y,hue,kind,col}. where x="fpr", y="tpr", hue="concept"
    Returns:
        None or pd.Dataframe, If models don't have ROC curve, return None
    """
    sns.set_palette("Paired")
    outs, comparator = self._loop_eval_handlers("roc_curve", **roc_curve_kwargs)
    all_dfs = []
    for _, (df, anchor) in enumerate(zip(outs, comparator)):
      df[self.comparator] = [anchor for _ in range(len(df))]
      all_dfs.append(df)

    if all_dfs:
      all_dfs = pd.concat(all_dfs, axis=0)
      if save_path or show:
        relplot_kwargs = self._setup_default_lineplot(all_dfs, relplot_kwargs)
        g = sns.relplot(
            data=all_dfs,
            x="fpr",
            y="tpr",
            hue='concept',
            kind="line",
            col=self.comparator,
            **relplot_kwargs)
        g.set_titles(col_template=str(self.comparator) + ':{col_name}', fontsize=5)
        if show:
          plt.show()
        if save_path:
          g.savefig(save_path)

    return all_dfs if isinstance(all_dfs, pd.DataFrame) else None

  def pr_plot(self,
              show=True,
              save_path: str = None,
              pr_curve_kwargs: dict = {},
              relplot_kwargs: dict = {}) -> Union[pd.DataFrame, None]:
    """Return dataframe of PR curve
    Args:
        show (bool, optional): Show the chart. Defaults to True.
        save_path (str): path to save rendered chart.
        pr_curve_kwargs (dict): keyword args of `eval_handler[...].model.pr_curve` method.
        relplot_kwargs (dict): keyword args of `sns.relplot` except {data,x,y,hue,kind,col} where x="recall", y="precision", hue="concept"
    Returns:
        None or pd.Dataframe, If models don't have PR curve, return None
    """
    sns.set_palette("Paired")
    outs, comparator = self._loop_eval_handlers("pr_curve", **pr_curve_kwargs)
    all_dfs = []
    for _, (df, anchor) in enumerate(zip(outs, comparator)):
      df[self.comparator] = [anchor for _ in range(len(df))]
      all_dfs.append(df)

    if all_dfs:
      all_dfs = pd.concat(all_dfs, axis=0)
      if save_path or show:
        relplot_kwargs = self._setup_default_lineplot(all_dfs, relplot_kwargs)
        g = sns.relplot(
            data=all_dfs,
            x="recall",
            y="precision",
            hue='concept',
            kind="line",
            col=self.comparator,
            **relplot_kwargs)
        g.set_titles(col_template=str(self.comparator) + ':{col_name}', fontsize=5)
        if show:
          plt.show()
        if save_path:
          g.savefig(save_path)

    return all_dfs if isinstance(all_dfs, pd.DataFrame) else None

  def all(
      self,
      output_folder: str,
      confidence_threshold: float = 0.5,
      iou_threshold: float = 0.5,
      overwrite: bool = False,
      metric_kwargs: dict = {},
      pr_plot_kwargs: dict = {},
      roc_plot_kwargs: dict = {},
  ):
    """Run all comparison methods one by one:
    - detailed_summary
    - pr_curve (if applicable)
    - pr_plot
    - confusion_matrix (if applicable)
    And save to output_folder

    Args:
      output_folder (str): path to output
      confidence_threshold (float): confidence threshold, applicable for classification and detection. Default is 0.5.
      iou_threshold (float): iou threshold, support in range(0.5, 1., step=0.1) applicable for detection.
      overwrite (bool): overwrite result of output_folder.
      metric_kwargs (dict): keyword args for `eval_handler[...].model.{method}`, except for {confidence_threshold, iou_threshold}.
      roc_plot_kwargs (dict): for relplot_kwargs of `roc_curve_plot` method.
      pr_plot_kwargs (dict): for relplot_kwargs of `pr_plot` method.
    """
    eval_type = get_eval_type(self.model_type)
    area = metric_kwargs.pop("area", "all")
    bypass_const = metric_kwargs.pop("bypass_const", False)

    fname = f"conf-{confidence_threshold}"
    if eval_type == EvalType.DETECTION:
      fname = f"{fname}_iou-{iou_threshold}_area-{area}"

    def join_root(*args):
      return os.path.join(output_folder, *args)

    output_folder = join_root(fname)
    if os.path.exists(output_folder) and not overwrite:
      raise RuntimeError(f"{output_folder} exists. If you want to overwrite, set `overwrite=True`")

    os.makedirs(output_folder, exist_ok=True)

    logger.info("Making summary tables...")
    dfs = self.detailed_summary(
        confidence_threshold=confidence_threshold,
        iou_threshold=iou_threshold,
        area=area,
        bypass_const=bypass_const)
    if dfs is not None:
      concept_df, total_df = dfs
      concept_df.to_csv(join_root("concepts_summary.csv"))
      total_df.to_csv(join_root("total_summary.csv"))

    curve_metric_kwargs = dict(
        confidence_threshold=confidence_threshold, iou_threshold=iou_threshold)
    curve_metric_kwargs.update(metric_kwargs)

    self.roc_curve_plot(
        show=False,
        save_path=join_root("roc.jpg"),
        roc_curve_kwargs=curve_metric_kwargs,
        relplot_kwargs=roc_plot_kwargs)

    self.pr_plot(
        show=False,
        save_path=join_root("pr.jpg"),
        pr_curve_kwargs=curve_metric_kwargs,
        relplot_kwargs=pr_plot_kwargs)

    self.confusion_matrix(
        show=False, save_path=join_root("confusion_matrix.jpg"), cm_kwargs=curve_metric_kwargs)

    logger.info(f"Done. Your outputs are saved at {output_folder}")



import importlib
import inspect
import os
from typing import Dict, Tuple, Type, Union

from clarifai_grpc.grpc.api.service_pb2 import MultiDatasetVersionMetricsGroupResponse
from google.protobuf.json_format import MessageToDict

from clarifai.constants.dataset import TASK_TO_ANNOTATION_TYPE
from clarifai.datasets.upload.base import ClarifaiDataLoader
from clarifai.errors import UserError


def load_module_dataloader(module_dir: Union[str, os.PathLike], **kwargs) -> ClarifaiDataLoader:
  """Validate and import dataset module data generator.
  Args:
    `module_dir`: relative path to the module directory
    The directory must contain a `dataset.py` script and the data itself.
    kwargs: keyword arguments to be passed to the dataloader class
  Module Directory Structure:
  ---------------------------
      <folder_name>/
      ├──__init__.py
      ├──<Your local dir dataset>/
      └──dataset.py
  dataset.py must implement a class named following the convention,
  <dataset_name>DataLoader and this class must inherit from base ClarifaiDataLoader()
  """
  module_path = os.path.join(module_dir, "dataset.py")
  spec = importlib.util.spec_from_file_location("dataset", module_path)

  if not spec:
    raise ImportError(f"Module not found at {module_path}")

  # Load the module using the spec
  dataset = importlib.util.module_from_spec(spec)
  # Execute the module to make its contents available
  spec.loader.exec_module(dataset)

  # get main module class
  main_module_cls = None
  for name, obj in dataset.__dict__.items():
    if inspect.isclass(obj) and "DataLoader" in name:
      main_module_cls = obj
    else:
      continue

  return main_module_cls(**kwargs)


class DisplayUploadStatus:
  """Class to display dataset upload status."""

  def __init__(self, dataloader: ClarifaiDataLoader,
               dataset_metrics_response: Type[MultiDatasetVersionMetricsGroupResponse],
               dataset_info_dict: Dict[str, str],
               pre_upload_stats: Tuple[Dict[str, int], Dict[str, int]]) -> None:
    """Initialize the class.
    Args:
      dataloader: ClarifaiDataLoader object
      dataset_metrics_response: The dataset version metrics response from the server.
      dataset_info_dict: The dataset info dictionary.
      pre_upload_stats: The pre upload stats for the dataset.
    """
    self.dataloader = dataloader
    self.dataset_metrics_response = dataset_metrics_response
    self.dataset_info_dict = dataset_info_dict
    self.pre_upload_stats = pre_upload_stats

    self.display()

  def display(self) -> None:
    """Display the upload status."""
    from rich.console import Console

    local_inputs_count, local_annotations_dict = self.get_dataloader_stats()
    uploaded_inputs_dict, uploaded_annotations_dict = self.get_dataset_version_stats(
        self.dataset_metrics_response)

    # Subtract the pre upload stats from the uploaded stats
    uploaded_inputs_dict = {
        key: int(uploaded_inputs_dict[key]) - int(self.pre_upload_stats[0].get(key, 0))
        for key in uploaded_inputs_dict
    }
    uploaded_annotations_dict = {
        key: uploaded_annotations_dict[key] - self.pre_upload_stats[1].get(key, 0)
        for key in uploaded_annotations_dict
    }

    self.local_annotations_count = sum(local_annotations_dict.values())
    self.uploaded_annotations_count = sum(uploaded_annotations_dict.values())

    local_dataset_dict = {
        "Inputs Count": str(local_inputs_count),
        "Annotations Count": str(local_annotations_dict)
    }
    uploaded_dataset_dict = {
        "Inputs Count": str(uploaded_inputs_dict["total"]),
        "Annotations Count": str(uploaded_annotations_dict)
    }

    panel_layout = self.get_display_layout(local_dataset_dict, uploaded_dataset_dict)

    console = Console()
    console.print(panel_layout)

  def get_dataloader_stats(self) -> Tuple[int, Dict[str, int]]:
    """Get the number of inputs and annotations in a dataloader.

    Returns:
      local_inputs_count (int): total number of inputs in the dataloader
      local_annotations_dict (Dict[str, int]): total number of annotations in the dataloader
    """
    from clarifai.constants.dataset import DATASET_UPLOAD_TASKS

    task = self.dataloader.task
    if task not in DATASET_UPLOAD_TASKS:
      raise UserError(
          "Invalid task, please use one of the following: {}".format(DATASET_UPLOAD_TASKS))
    local_inputs_count = len(self.dataloader)
    local_annotations_dict = dict(concepts=0, bboxes=0, polygons=0)
    for i in range(local_inputs_count):
      key, attr = [(k, v) for k, v in TASK_TO_ANNOTATION_TYPE.get(task).items()][0]
      local_annotations_dict[key] += len(getattr(self.dataloader[i], attr))
    return local_inputs_count, local_annotations_dict

  @staticmethod
  def get_dataset_version_stats(
      dataset_metrics_response: Type[MultiDatasetVersionMetricsGroupResponse]
  ) -> Tuple[Dict[str, int], Dict[str, int]]:
    """Parse the response from the server for the dataset version metrics groups.
    Args:
      dataset_metrics_response: The dataset version metrics response from the server.

    Returns:
      uploaded_inputs_dict (Dict[str, int]): The input statistics for the dataset.
      uploaded_annotations_dict (Dict[str, int]): The annotation statistics for the dataset.
    """
    dataset_statistics = []
    uploaded_inputs_dict = {}
    uploaded_annotations_dict = dict(concepts=0, bboxes=0, polygons=0)
    dict_response = MessageToDict(dataset_metrics_response)

    for data in dict_response["datasetVersionMetricsGroups"]:
      if isinstance(data["value"], str):
        if data["value"].startswith("id-"):
          data["metrics"].update({"Concept": data["value"]})
          data["metrics"].pop("regionLocationMatrix", None)
          dataset_statistics.append(data["metrics"])
        else:
          uploaded_inputs_dict[data["value"]] = data["metrics"]["inputsCount"]

    for ds in dataset_statistics:
      uploaded_annotations_dict["bboxes"] += int(ds["boundingBoxesCount"])
      uploaded_annotations_dict["concepts"] += int(ds["positiveInputTagsCount"])
      uploaded_annotations_dict["polygons"] += int(ds["polygonsCount"])

    return uploaded_inputs_dict, uploaded_annotations_dict

  def _create_layout(self):
    from rich.layout import Layout
    from rich.progress import BarColumn, Progress, TextColumn

    # Create a Layout
    layout = Layout()

    # Add a new task to the progress bar
    progress = Progress(
        "{task.description}",
        BarColumn(bar_width=100),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    )

    # Split the layout into top and bottom rows
    layout.split(Layout(name="Progress"), Layout(name="Tables"))

    # Add the progress bar to the top layout
    layout["Progress"].update(progress)

    # Split the bottom layout into two columns
    layout["Tables"].split_row(
        Layout(name="Local Dataset"),
        Layout(name="Uploaded Dataset"),
    )

    # Create a new layout for the panels
    panel_layout = Layout(size=18)

    # Split the panel layout into top and bottom rows
    panel_layout.split(Layout(name="Progress Panel", size=9), Layout(name="Tables Panel", size=9))

    return layout, panel_layout, progress

  def get_display_layout(self, local_dataset_dict: Dict[str, str],
                         uploaded_dataset_dict: Dict[str, str]):
    """Create a layout for the display.

    Args:
      local_dataset_dict (dict): The local dataset stats info dict.
      uploaded_dataset_dict (dict): The uploaded dataset stats info dict.

    Returns:
      panel_layout (Layout): The panel layout for the display.
    """
    from rich.console import Group
    from rich.panel import Panel

    from clarifai.utils.logging import table_from_dict

    local_dataset_table = table_from_dict(
        [local_dataset_dict],
        column_names=["Inputs Count", "Annotations Count"],
        title="[cyan]Local Dataset")
    uploaded_dataset_table = table_from_dict(
        [uploaded_dataset_dict],
        column_names=["Inputs Count", "Annotations Count"],
        title="[cyan]Uploaded Dataset")
    dataset_info_table = table_from_dict(
        [self.dataset_info_dict], column_names=["dataset_id", "user_id", "app_id"])

    layout, panel_layout, progress = self._create_layout()

    # Add a new task to the progress bar
    progress.add_task(
        "[cyan]Inputs Progress:",
        completed=int(uploaded_dataset_dict["Inputs Count"]),
        total=int(local_dataset_dict["Inputs Count"]))
    progress.add_task(
        "[cyan]Annotations Progress:",
        completed=self.uploaded_annotations_count,
        total=self.local_annotations_count)

    # Add the tables to the respective layouts
    layout["Local Dataset"].update(local_dataset_table)
    layout["Uploaded Dataset"].update(uploaded_dataset_table)

    # Create a render group for the progress bar and the additional data
    progress_group = Group(progress, dataset_info_table)

    # Create a panel for the progress bar with a blue border and a suitable heading
    progress_panel = Panel(progress_group, title="[b] Dataset Upload Summary", border_style="blue")

    # Create a panel for the tables comparison with a blue border and a suitable heading
    tables_panel = Panel(
        layout["Tables"],
        title="[b] Dataset Metrics Comparison",
        border_style="blue",
        expand=False)

    # Add the panels to the respective layouts
    panel_layout["Progress Panel"].update(progress_panel)
    panel_layout["Tables Panel"].update(tables_panel)

    return panel_layout



from schema import And, Optional, Regex, Schema


def get_schema() -> Schema:
  """Initialize the schema for rank and filter.

        This schema validates:

        - Rank and filter must be a list
        - Each item in the list must be a dict
        - The dict can contain these optional keys:
            - 'image_url': Valid URL string
            - 'text_raw': Non-empty string
            - 'metadata': Dict
            - 'image_bytes': Bytes
            - 'geo_point': Dict with 'longitude', 'latitude' and 'geo_limit' as float, float and int respectively
            - 'concepts': List where each item is a concept dict
        - Concept dict requires at least one of:
            - 'name': Non-empty string with dashes/underscores
            - 'id': Non-empty string
            - 'language': Non-empty string
            - 'value': 0 or 1 integer
        - 'input_types': List of 'image', 'video', 'text' or 'audio'
        - 'input_dataset_ids': List of strings
        - 'input_status_code': Integer

        Returns:
            Schema: The schema for rank and filter.
        """
  # Schema for a single concept
  concept_schema = Schema({
      Optional('value'):
          And(int, lambda x: x in [0, 1]),
      Optional('id'):
          And(str, len),
      Optional('language'):
          And(str, len),
      # Non-empty strings with internal dashes and underscores.
      Optional('name'):
          And(str, len, Regex(r'^[0-9A-Za-z]+([-_][0-9A-Za-z]+)*$'))
  })

  # Schema for a rank or filter item
  rank_filter_item_schema = Schema({
      Optional('image_url'):
          And(str, Regex(r'^https?://')),
      Optional('text_raw'):
          And(str, len),
      Optional('metadata'):
          dict,
      Optional('image_bytes'):
          bytes,
      Optional('geo_point'): {
          'longitude': float,
          'latitude': float,
          'geo_limit': int
      },
      Optional("concepts"):
          And(list, lambda x: all(concept_schema.is_valid(item) and len(item) > 0 for item in x)),

      ## input filters
      Optional('input_types'):
          And(list, lambda input_types: all(input_type in ('image', 'video', 'text', 'audio')
                                            for input_type in input_types)),
      Optional('input_dataset_ids'):
          list,
      Optional('input_status_code'):
          int,
  })

  # Schema for rank and filter args
  return Schema([rank_filter_item_schema])



#! COCO detection dataset

import os

from ..base import ClarifaiDataLoader

from ..features import VisualDetectionFeatures

#pycocotools is a dependency for this loader
try:
  from pycocotools.coco import COCO
except ImportError:
  raise ImportError("Could not import pycocotools package. "
                    "Please do `pip install 'clarifai[all]'` to import pycocotools.")


class COCODetectionDataLoader(ClarifaiDataLoader):

  def __init__(self, images_dir, label_filepath):
    """
      Args:
        images_dir: Directory containing the images.
        label_filepath: Path to the COCO annotation file.
      """
    self.images_dir = images_dir
    self.label_filepath = label_filepath

    self.map_ids = {}
    self.load_data()

  @property
  def task(self):
    return "visual_detection"

  def load_data(self) -> None:
    self.coco = COCO(self.label_filepath)
    self.map_ids = {i: img_id for i, img_id in enumerate(list(self.coco.imgs.keys()))}

  def __getitem__(self, index: int):
    value = self.coco.imgs[self.map_ids[index]]
    image_path = os.path.join(self.images_dir, value['file_name'])
    annots = []  # bboxes
    concept_ids = []

    input_ann_ids = self.coco.getAnnIds(imgIds=[value['id']])
    input_anns = self.coco.loadAnns(input_ann_ids)

    for ann in input_anns:
      # get concept info
      # note1: concept_name can be human readable
      # note2: concept_id can only be alphanumeric, up to 32 characters, with no special chars except `-` and `_`
      concept_name = self.coco.cats[ann['category_id']]['name']
      concept_id = concept_name.lower().replace(' ', '-')

      # get bbox information
      # note1: coco bboxes are `[x_min, y_min, width, height]` in pixels
      # note2: clarifai bboxes are `[x_min, y_min, x_max, y_max]` normalized between 0-1.0
      coco_bbox = ann['bbox']
      clarifai_bbox = {
          'left_col': max(0, coco_bbox[0] / value['width']),
          'top_row': max(0, coco_bbox[1] / value['height']),
          'right_col': min(1, (coco_bbox[0] + coco_bbox[2]) / value['width']),
          'bottom_row': min(1, (coco_bbox[1] + coco_bbox[3]) / value['height'])
      }
      if (clarifai_bbox['left_col'] >=
          clarifai_bbox['right_col']) or (clarifai_bbox['top_row'] >= clarifai_bbox['bottom_row']):
        continue
      annots.append([
          clarifai_bbox['left_col'], clarifai_bbox['top_row'], clarifai_bbox['right_col'],
          clarifai_bbox['bottom_row']
      ])
      concept_ids.append(concept_id)

    assert len(concept_ids) == len(annots), f"Num concepts must match num bbox annotations\
        for a single image. Found {len(concept_ids)} concepts and {len(annots)} bboxes."

    return VisualDetectionFeatures(image_path, concept_ids, annots, id=str(value['id']))

  def __len__(self):
    return len(self.coco.imgs)



from datetime import datetime
from typing import Any, Callable

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf import struct_pb2
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.wrappers_pb2 import BoolValue

from clarifai.client.auth import create_stub
from clarifai.client.auth.helper import ClarifaiAuthHelper
from clarifai.constants.base import COMPUTE_ORCHESTRATION_RESOURCES
from clarifai.errors import ApiError, UserError
from clarifai.utils.constants import CLARIFAI_PAT_ENV_VAR, CLARIFAI_SESSION_TOKEN_ENV_VAR
from clarifai.utils.misc import get_from_dict_or_env


class BaseClient:
  """BaseClient is the base class for all the classes interacting with Clarifai endpoints.

  Args:
      **kwargs: Additional keyword arguments to be passed to the ClarifaiAuthHelper.
          - user_id (str): A user ID for authentication.
          - app_id (str): An app ID for the application to interact with.
          - pat (str): A personal access token for authentication.
          - token (str): A session token for authentication. Accepts either a session token or a pat.
          - base (str): The base URL for the API endpoint. Defaults to 'https://api.clarifai.com'.
          - ui (str): The URL for the UI. Defaults to 'https://clarifai.com'.
          - root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.


  Attributes:
      auth_helper (ClarifaiAuthHelper): An instance of ClarifaiAuthHelper for authentication.
      STUB (Stub): The gRPC Stub object for API interaction.
      metadata (tuple): The gRPC metadata containing the personal access token.
      userDataObject (UserAppIDSet): The protobuf object representing user and app IDs.
      base (str): The base URL for the API endpoint.
  """

  def __init__(self, **kwargs):
    token, pat = "", ""
    try:
      pat = get_from_dict_or_env(key="pat", env_key=CLARIFAI_PAT_ENV_VAR, **kwargs)
    except UserError:
      token = get_from_dict_or_env(key="token", env_key=CLARIFAI_SESSION_TOKEN_ENV_VAR, **kwargs)
    finally:
      assert token or pat, Exception(
          "Need 'pat' or 'token' in args or use one of the CLARIFAI_PAT or CLARIFAI_SESSION_TOKEN env vars"
      )
    kwargs.update({'token': token, 'pat': pat})

    self.auth_helper = ClarifaiAuthHelper(**kwargs, validate=False)
    self.STUB = create_stub(self.auth_helper)
    self.metadata = self.auth_helper.metadata
    self.pat = self.auth_helper.pat
    self.token = self.auth_helper._token
    self.user_app_id = self.auth_helper.get_user_app_id_proto()
    self.base = self.auth_helper.base
    self.root_certificates_path = self.auth_helper._root_certificates_path

  @classmethod
  def from_env(cls, validate: bool = False):
    auth = ClarifaiAuthHelper.from_env(validate=validate)
    return cls.from_auth_helper(auth)

  @classmethod
  def from_auth_helper(cls, auth: ClarifaiAuthHelper, **kwargs):
    default_kwargs = {
        "user_id":
            kwargs.get("user_id", None) or auth.user_id,
        "app_id":
            kwargs.get("app_id", None) or auth.app_id,
        "pat":
            kwargs.get("pat", None) or auth.pat,
        "token":
            kwargs.get("token", None) or auth._token,
        "root_certificates_path":
            kwargs.get("root_certificates_path", None) or auth._root_certificates_path
    }
    _base = kwargs.get("base", None) or auth.base
    _clss = cls.__mro__[0]
    if _clss == BaseClient:
      kwargs = {
        **default_kwargs,
        "base": _base, # Baseclient uses `base`
        "ui": kwargs.get("ui", None) or auth.ui
      }
    else:
      # Remove user_id and app_id if a custom URL is provided
      if kwargs.get("url"):
        default_kwargs.pop("user_id", "")
        default_kwargs.pop("app_id", "")
      # Remove app_id if the class name is a compute orchestration resource
      if any(co_resource in _clss.__name__ for co_resource in COMPUTE_ORCHESTRATION_RESOURCES):
        default_kwargs.pop("app_id", "")
      kwargs.update({**default_kwargs, "base_url": _base})

    return cls(**kwargs)

  def _grpc_request(self, method: Callable, argument: Any):
    """Makes a gRPC request to the API.

    Args:
        method (Callable): The gRPC method to call.
        argument (Any): The argument to pass to the gRPC method.

    Returns:
        res (Any): The result of the gRPC method call.
    """

    try:
      res = method(argument, metadata=self.auth_helper.metadata)
      # MessageToDict(res) TODO global debug logger
      return res
    except ApiError:
      raise Exception("ApiError")

  def convert_string_to_timestamp(self, date_str) -> Timestamp:
    """Converts a string to a Timestamp object.

    Args:
        date_str (str): The string to convert.

    Returns:
        Timestamp: The converted Timestamp object.
    """
    # Parse the string into a Python datetime object
    try:
      datetime_obj = datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%S.%fZ')
    except ValueError:
      try:
        datetime_obj = datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%SZ')
      except ValueError:
        return Timestamp()

    # Convert the datetime object to a Timestamp object
    timestamp_obj = Timestamp()
    timestamp_obj.FromDatetime(datetime_obj)

    return timestamp_obj

  def process_response_keys(self, old_dict, listing_resource=None):
    """Converts keys in a response dictionary to resource proto format.

    Args:
        old_dict (dict): The dictionary to convert.

    Returns:
        new_dict (dict): The dictionary with processed keys.
    """
    if listing_resource:
      old_dict[f'{listing_resource}_id'] = old_dict['id']
      old_dict.pop('id')

    def convert_recursive(item):
      if isinstance(item, dict):
        new_item = {}
        for key, value in item.items():
          if key == 'default_value':
            # Map infer param value to proto value
            value_map = dict(number_value=None, string_value=None, bool_value=None)

            def map_fn(v):
              return 'number_value' if isinstance(v, float) or isinstance(v, int) else \
              'string_value' if isinstance(v, str) else \
              'bool_value' if isinstance(v, bool) else None

            value_map[map_fn(value)] = value
            value = struct_pb2.Value(**value_map)
          if key in ['created_at', 'modified_at', 'completed_at']:
            value = self.convert_string_to_timestamp(value)
          elif key in ['workflow_recommended', 'is_template']:
            value = BoolValue(value=True)
          elif key in ['fields_map', 'params']:
            value_s = struct_pb2.Struct()
            value_s.update(value)
            value = value_s
          elif key == 'metrics':
            continue
          elif key == 'size':
            value = int(value)
          elif key == 'image_info':
            value = resources_pb2.ImageInfo(**value)
          elif key == 'hosted_image_info':
            continue
          elif key in ['metadata', 'presets']:
            if isinstance(value, dict) and value != {}:
              value_s = struct_pb2.Struct()
              value_s.update(value)
              value = value_s
            else:
              continue
          new_item[key] = convert_recursive(value)
        return new_item
      elif isinstance(item, list):
        return [convert_recursive(element) for element in item]
      else:
        return item

    new_dict = convert_recursive(old_dict)
    return new_dict



import argparse
import os
import time
from string import Template

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf import json_format
from rich import print

from clarifai.client import BaseClient
from clarifai.runners.utils.loader import HuggingFaceLoarder
from clarifai.utils.logging import logger


def _clear_line(n: int = 1) -> None:
  LINE_UP = '\033[1A'  # Move cursor up one line
  LINE_CLEAR = '\x1b[2K'  # Clear the entire line
  for _ in range(n):
    print(LINE_UP, end=LINE_CLEAR, flush=True)


class ModelUploader:
  DEFAULT_PYTHON_VERSION = 3.11
  CONCEPTS_REQUIRED_MODEL_TYPE = [
      'visual-classifier', 'visual-detector', 'visual-segmenter', 'text-classifier'
  ]

  def __init__(self, folder: str):
    self._client = None
    self.folder = self._validate_folder(folder)
    self.config = self._load_config(os.path.join(self.folder, 'config.yaml'))
    self.model_proto = self._get_model_proto()
    self.model_id = self.model_proto.id
    self.inference_compute_info = self._get_inference_compute_info()
    self.is_v3 = True  # Do model build for v3

  @staticmethod
  def _validate_folder(folder):
    if not folder.startswith("/"):
      folder = os.path.join(os.getcwd(), folder)
    logger.info(f"Validating folder: {folder}")
    if not os.path.exists(folder):
      raise FileNotFoundError(f"Folder {folder} not found, please provide a valid folder path")
    files = os.listdir(folder)
    assert "requirements.txt" in files, "requirements.txt not found in the folder"
    assert "config.yaml" in files, "config.yaml not found in the folder"
    assert "1" in files, "Subfolder '1' not found in the folder"
    subfolder_files = os.listdir(os.path.join(folder, '1'))
    assert 'model.py' in subfolder_files, "model.py not found in the folder"
    return folder

  @staticmethod
  def _load_config(config_file: str):
    with open(config_file, 'r') as file:
      config = yaml.safe_load(file)
    return config

  @property
  def client(self):
    if self._client is None:
      assert "model" in self.config, "model info not found in the config file"
      model = self.config.get('model')
      assert "user_id" in model, "user_id not found in the config file"
      assert "app_id" in model, "app_id not found in the config file"
      user_id = model.get('user_id')
      app_id = model.get('app_id')

      base = os.environ.get('CLARIFAI_API_BASE', 'https://api-dev.clarifai.com')

      self._client = BaseClient(user_id=user_id, app_id=app_id, base=base)
      logger.info(f"Client initialized for user {user_id} and app {app_id}")
    return self._client

  def _get_model_proto(self):
    assert "model" in self.config, "model info not found in the config file"
    model = self.config.get('model')

    assert "model_type_id" in model, "model_type_id not found in the config file"
    assert "id" in model, "model_id not found in the config file"
    assert "user_id" in model, "user_id not found in the config file"
    assert "app_id" in model, "app_id not found in the config file"

    model_proto = json_format.ParseDict(model, resources_pb2.Model())
    assert model_proto.id == model_proto.id.lower(), "Model ID must be lowercase"
    assert model_proto.user_id == model_proto.user_id.lower(), "User ID must be lowercase"
    assert model_proto.app_id == model_proto.app_id.lower(), "App ID must be lowercase"

    return model_proto

  def _get_inference_compute_info(self):
    assert ("inference_compute_info" in self.config
           ), "inference_compute_info not found in the config file"
    inference_compute_info = self.config.get('inference_compute_info')
    return json_format.ParseDict(inference_compute_info, resources_pb2.ComputeInfo())

  def maybe_create_model(self):
    resp = self.client.STUB.GetModel(
        service_pb2.GetModelRequest(
            user_app_id=self.client.user_app_id, model_id=self.model_proto.id))
    if resp.status.code == status_code_pb2.SUCCESS:
      logger.info(
          f"Model '{self.client.user_app_id.user_id}/{self.client.user_app_id.app_id}/models/{self.model_proto.id}' already exists, "
          f"will create a new version for it.")
      return resp

    request = service_pb2.PostModelsRequest(
        user_app_id=self.client.user_app_id,
        models=[self.model_proto],
    )
    return self.client.STUB.PostModels(request)

  def create_dockerfile(self):
    num_accelerators = self.inference_compute_info.num_accelerators
    if num_accelerators:
      dockerfile_template = os.path.join(
          os.path.dirname(os.path.dirname(__file__)),
          'dockerfile_template',
          'Dockerfile.cuda.template',
      )
    else:
      dockerfile_template = os.path.join(
          os.path.dirname(os.path.dirname(__file__)), 'dockerfile_template',
          'Dockerfile.cpu.template')

    with open(dockerfile_template, 'r') as template_file:
      dockerfile_template = template_file.read()

    dockerfile_template = Template(dockerfile_template)

    # Get the Python version from the config file
    build_info = self.config.get('build_info', {})
    if 'python_version' in build_info:
      python_version = build_info['python_version']
      logger.info(
          f"Using Python version {python_version} from the config file to build the Dockerfile")
    else:
      logger.info(
          f"Python version not found in the config file, using default Python version: {self.DEFAULT_PYTHON_VERSION}"
      )
      python_version = self.DEFAULT_PYTHON_VERSION

    # Replace placeholders with actual values
    dockerfile_content = dockerfile_template.safe_substitute(
        PYTHON_VERSION=python_version,
        name='main',
    )

    # Write Dockerfile
    with open(os.path.join(self.folder, 'Dockerfile'), 'w') as dockerfile:
      dockerfile.write(dockerfile_content)

  def download_checkpoints(self):
    if not self.config.get("checkpoints"):
      logger.info("No checkpoints specified in the config file")
      return

    assert "type" in self.config.get("checkpoints"), "No loader type specified in the config file"
    loader_type = self.config.get("checkpoints").get("type")
    if not loader_type:
      logger.info("No loader type specified in the config file for checkpoints")
    assert loader_type == "huggingface", "Only huggingface loader supported for now"
    if loader_type == "huggingface":
      assert "repo_id" in self.config.get("checkpoints"), "No repo_id specified in the config file"
      repo_id = self.config.get("checkpoints").get("repo_id")

      # prefer env var for HF_TOKEN but if not provided then use the one from config.yaml if any.
      if 'HF_TOKEN' in os.environ:
        hf_token = os.environ['HF_TOKEN']
      else:
        hf_token = self.config.get("checkpoints").get("hf_token", None)
        assert hf_token != 'hf_token', "The default 'hf_token' is not valid. Please provide a valid token or leave that field out of config.yaml if not needed."
      loader = HuggingFaceLoarder(repo_id=repo_id, token=hf_token)

      checkpoint_path = os.path.join(self.folder, '1', 'checkpoints')
      success = loader.download_checkpoints(checkpoint_path)

      if not success:
        logger.error(f"Failed to download checkpoints for model {repo_id}")
        return
      logger.info(f"Downloaded checkpoints for model {repo_id}")

  def _concepts_protos_from_concepts(self, concepts):
    concept_protos = []
    for concept in concepts:
      concept_protos.append(resources_pb2.Concept(
          id=str(concept[0]),
          name=concept[1],
      ))
    return concept_protos

  def hf_labels_to_config(self, labels, config_file):
    with open(config_file, 'r') as file:
      config = yaml.safe_load(file)
    model = config.get('model')
    model_type_id = model.get('model_type_id')
    assert model_type_id in self.CONCEPTS_REQUIRED_MODEL_TYPE, f"Model type {model_type_id} not supported for concepts"
    concept_protos = self._concepts_protos_from_concepts(labels)

    config['concepts'] = [{'id': concept.id, 'name': concept.name} for concept in concept_protos]

    with open(config_file, 'w') as file:
      yaml.dump(config, file, sort_keys=False)
    concepts = config.get('concepts')
    logger.info(f"Updated config.yaml with {len(concepts)} concepts.")

  def _get_model_version_proto(self):

    model_version = resources_pb2.ModelVersion(
        pretrained_model_config=resources_pb2.PretrainedModelConfig(),
        inference_compute_info=self.inference_compute_info,
    )

    model_type_id = self.config.get('model').get('model_type_id')
    if model_type_id in self.CONCEPTS_REQUIRED_MODEL_TYPE:

      loader = HuggingFaceLoarder()
      checkpoint_path = os.path.join(self.folder, '1', 'checkpoints')
      labels = loader.fetch_labels(checkpoint_path)
      # sort the concepts by id and then update the config file
      labels = sorted(labels.items(), key=lambda x: int(x[0]))

      config_file = os.path.join(self.folder, 'config.yaml')
      self.hf_labels_to_config(labels, config_file)

      model_version.output_info.data.concepts.extend(self._concepts_protos_from_concepts(labels))
    return model_version

  def upload_model_version(self):
    file_path = f"{self.folder}.tar.gz"
    logger.info(f"Will tar it into file: {file_path}")

    # Tar the folder
    os.system(f"tar --exclude=*~ -czvf {self.folder}.tar.gz -C {self.folder} .")
    logger.info("Tarring complete, about to start upload.")

    model_version = self._get_model_version_proto()

    response = self.maybe_create_model()

    for response in self.client.STUB.PostModelVersionsUpload(
        self.model_version_stream_upload_iterator(model_version, file_path),):
      percent_completed = 0
      if response.status.code == status_code_pb2.UPLOAD_IN_PROGRESS:
        percent_completed = response.status.percent_completed
      details = response.status.details

      _clear_line()
      print(
          f"Status: {response.status.description}, "
          f"Progress: {percent_completed}% - {details} ",
          f"request_id: {response.status.req_id}",
          end='\r',
          flush=True)
    print()
    if response.status.code != status_code_pb2.MODEL_BUILDING:
      logger.error(f"Failed to upload model version: {response.status.description}")
      return
    model_version_id = response.model_version_id
    logger.info(f"Created Model Version ID: {model_version_id}")

    self.monitor_model_build(model_version_id)

  def model_version_stream_upload_iterator(self, model_version, file_path):
    yield self.init_upload_model_version(model_version, file_path)
    with open(file_path, "rb") as f:
      file_size = os.path.getsize(file_path)
      chunk_size = int(127 * 1024 * 1024)  # 127MB chunk size
      num_chunks = (file_size // chunk_size) + 1
      logger.info("Uploading file...")
      logger.info("File size: ", file_size)
      logger.info("Chunk size: ", chunk_size)
      logger.info("Number of chunks: ", num_chunks)
      read_so_far = 0
      for part_id in range(num_chunks):
        try:
          chunk_size = min(chunk_size, file_size - read_so_far)
          chunk = f.read(chunk_size)
          if not chunk:
            break
          read_so_far += len(chunk)
          yield service_pb2.PostModelVersionsUploadRequest(
              content_part=resources_pb2.UploadContentPart(
                  data=chunk,
                  part_number=part_id + 1,
                  range_start=read_so_far,
              ))
        except Exception as e:
          logger.exception(f"\nError uploading file: {e}")
          break

    if read_so_far == file_size:
      logger.info("\nUpload complete!, waiting for model build...")

  def init_upload_model_version(self, model_version, file_path):
    file_size = os.path.getsize(file_path)
    logger.info(f"Uploading model version '{model_version.id}' of model {self.model_proto.id}")
    logger.info(f"Using file '{os.path.basename(file_path)}' of size: {file_size} bytes")
    return service_pb2.PostModelVersionsUploadRequest(
        upload_config=service_pb2.PostModelVersionsUploadConfig(
            user_app_id=self.client.user_app_id,
            model_id=self.model_proto.id,
            model_version=model_version,
            total_size=file_size,
            is_v3=self.is_v3,
        ))

  def monitor_model_build(self, model_version_id):
    st = time.time()
    while True:
      resp = self.client.STUB.GetModelVersion(
          service_pb2.GetModelVersionRequest(
              user_app_id=self.client.user_app_id,
              model_id=self.model_proto.id,
              version_id=model_version_id,
          ))
      status_code = resp.model_version.status.code
      if status_code == status_code_pb2.MODEL_BUILDING:
        logger.info(
            f"Model is building... (elapsed {time.time() - st:.1f}s)", end='\r', flush=True)
        time.sleep(1)
      elif status_code == status_code_pb2.MODEL_TRAINED:
        logger.info("\nModel build complete!")
        logger.info(
            f"Check out the model at https://clarifai.com/{self.client.user_app_id.user_id}/apps/{self.client.user_app_id.app_id}/models/{self.model_id}/versions/{model_version_id}"
        )
        break
      else:
        logger.info(
            f"\nModel build failed with status: {resp.model_version.status} and response {resp}")
        break


def main(folder, download_checkpoints):
  uploader = ModelUploader(folder)
  if download_checkpoints:
    uploader.download_checkpoints()
  uploader.create_dockerfile()
  input("Press Enter to continue...")
  uploader.upload_model_version()


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--model_path', type=str, help='Path of the model folder to upload', required=True)
  # flag to default to not download checkpoints
  parser.add_argument(
      '--download_checkpoints',
      action='store_true',
      help=
      'Flag to download checkpoints before uploading and including them in the tar file that is uploaded. Defaults to False, which will attempt to download them at docker build time.',
  )
  args = parser.parse_args()

  main(args.model_path, args.download_checkpoints)



import os
import urllib.request
from typing import Any, Dict
from urllib.parse import urlparse

from clarifai_grpc.channel.clarifai_channel import ClarifaiChannel
from clarifai_grpc.grpc.api import resources_pb2, service_pb2_grpc
from clarifai.utils.constants import CLARIFAI_PAT_ENV_VAR, CLARIFAI_SESSION_TOKEN_ENV_VAR

DEFAULT_BASE = "https://api.clarifai.com"
DEFAULT_UI = "https://clarifai.com"

# Map from base domain to True / False for whether the base has https or http.
# This is filled in get_stub() if it's not in there already.
base_https_cache = {}
ui_https_cache = {}


def clear_cache() -> None:
  """Clears the cache."""
  base_https_cache.clear()
  ui_https_cache.clear()


def https_cache(cache: dict, url: str) -> str:
  """This is a helper function to cache whether a url is http or https."""
  HTTPS = True
  HTTP = False

  # If http or https is provided, we trust that it is correct.
  # Note: this always stores the url without http:// or https://
  if url.startswith("https://"):
    url = url.replace("https://", "")
    cache[url] = HTTPS
  elif url.startswith("http://"):
    url = url.replace("http://", "")
    cache[url] = HTTP
  elif url not in cache:
    # We know our endpoints are https.
    host_name = urlparse(url).hostname
    if host_name and host_name.endswith(".clarifai.com"):
      cache[url] = HTTPS
    else:  # need to test it.
      try:  # make request to https endpoint.
        urllib.request.urlopen("https://%s/v2/auth/methods" % url, timeout=1)
        cache[url] = HTTPS  # cache it.
      except Exception as e:
        if "SSL" in str(e):  # if ssl error then we know it's http.
          cache[url] = HTTP
          # For http urls we need host:port format.
          if ":" not in url:
            raise Exception("When providing an insecure url it must have both host:port format")
        else:
          raise Exception("Could not get a valid response from url: %s, is the API running there?"
                          % url) from e
  return url


class ClarifaiAuthHelper:

  def __init__(self,
               user_id: str = "",
               app_id: str = "",
               pat: str = "",
               token: str = "",
               base: str = DEFAULT_BASE,
               ui: str = DEFAULT_UI,
               root_certificates_path: str = None,
               validate: bool = True):
    """
        A helper to get the authorization information needed to make API calls with the grpc
        client to a specific app using a personal access token.

        There are classmethods to make this object easily from either query_params provided by streamlit or from env vars.

        Note: only one of personal access token (pat) or a session token (token) can be provided.
        Always use PATs in your code and never session tokens, those are only provided internal UI code.

    Args:
      user_id: a user id who owns the resource you want to make calls to.
      app_id: an app id for the application that owns the resource you want to interact with
      pat: a personal access token.
      token: a session token (internal use only, always use a PAT).
      base: a url to the API endpoint to hit. Examples include api.clarifai.com,
    https://api.clarifai.com (default), https://host:port, http://host:port, host:port (will be treated as http, not https). It's highly recommended to include the http:// or https:// otherwise we need to check the endpoint to determine if it has SSL during this __init__
      ui: a url to the UI. Examples include clarifai.com,
    https://clarifai.com (default), https://host:port, http://host:port, host:port (will be treated as http, not https). It's highly recommended to include the http:// or https:// otherwise we need to check the endpoint to determine if it has SSL during this __init__
      root_certificates_path: path to the root certificates file. This is only used for grpc secure channels.
      validate: whether to validate the inputs. This is useful for overriding vars then validating
    """

    self.user_id = user_id
    self.app_id = app_id
    self._pat = pat
    self._token = token
    self._root_certificates_path = root_certificates_path

    self.set_base(base)
    self.set_ui(ui)
    if validate:
      self.validate()

  def validate(self):
    if self.user_id == "":
      raise Exception(
          "Need 'user_id' to not be empty in the query params or user CLARIFAI_USER_ID env var")
    if self._pat != "" and self._token != "":
      raise Exception(
          "A personal access token OR a session token need to be provided, but you cannot provide both."
      )
    elif self._pat == "" and self._token == "":
      raise Exception(
          "Need 'pat' or 'token' in the query params or use one of the CLARIFAI_PAT or CLARIFAI_SESSION_TOKEN env vars"
      )
    if (self._root_certificates_path) and (not os.path.exists(self._root_certificates_path)):
      raise Exception("Root certificates path %s does not exist" % self._root_certificates_path)

  @classmethod
  def from_streamlit(cls, st: Any) -> "ClarifaiAuthHelper":
    """ This is a convenient method to check the environment variables first to see if there are
    required variables for auth, then override them with any additional query parameters that may
    have been passed in.

    Note: if a .streamlit/secrets.toml is present then st.secrets will auto populate the
    corresponding environment variables and we will pick them up from there, OVERWRITING whatever
    matching env var that may already be present.

        Args:
          st: the streamlit package typically as: 'import streamlit as st'
        Returns:
          auth: this class instantiated
        """
    # start with the env vars (potentially loaded from secrets.toml)
    # Don't validate yet as we'll layer on the query params next.
    auth = ClarifaiAuthHelper.from_env(validate=False)

    # Then add in the query params.
    try:
      auth.add_streamlit_query_params(st.experimental_get_query_params())
    except Exception as e:
      st.error(e)
      st.stop()
      raise e

    # Then validate.
    try:
      auth.validate()
    except Exception as e:
      st.error(e)
      st.stop()
      raise e

    return auth

  @classmethod
  def from_streamlit_query_params(cls, query_params: Any = "") -> "ClarifaiAuthHelper":
    """Initialize from streamlit queryparams. The following things will be looked for:
          user_id: as 'user_id' in query_params
          app_id: as 'app_id' in query_params
          one of:
            token: as 'token' in query_params
            pat: as 'pat' in query_params
          optionally:
            base: as 'base' in query_params.
            ui: as 'ui' in query_params.

    """

    # Setup an empty one (not from env).
    auth = ClarifaiAuthHelper("", "", "", "", validate=False)

    # Then add in the query params.
    auth.add_streamlit_query_params(query_params)

    # Then validate.
    auth.validate()

    return auth

  def add_streamlit_query_params(self, query_params: Any = "") -> "ClarifaiAuthHelper":
    """Initialize from streamlit queryparams. The following things will be looked for:
          user_id: as 'user_id' in query_params
          app_id: as 'app_id' in query_params
          one of:
            token: as 'token' in query_params
            pat: as 'pat' in query_params
          optionally:
            base: as 'base' in query_params.

    Args:
      query_params: the streamlit.experimental_get_query_params() response or an empty dict to fall
    back to using env vars.
    """
    error_description = """
Please check the following required query params are in the url:
 - 'user_id': the user ID accessing the module.
 - 'app_id': the app the module is being accessed from.
 - 'token' or 'pat': to authenticate the calling user with a session token or personal access token.

Additionally, these optional params are supported:
 - 'base': the base domain for the API such as https://api.clarifai.com
 - 'ui': the overall UI domain for redirects such as https://clarifai.com
"""

    if query_params == "":  # empty response from streamlit
      query_params = {}
    for k in ["user_id", "app_id", "token", "pat"]:
      if k in query_params and len(query_params[k]) != 1:
        err_str = "There should only be 1 query param value for key '%s'" % k
        raise Exception(err_str + error_description)
    if "user_id" in query_params:
      self.user_id = query_params["user_id"][0]
    if "app_id" in query_params:
      self.app_id = query_params["app_id"][0]
    if "token" in query_params:
      self._token = query_params["token"][0]
    if "pat" in query_params:
      self._pat = query_params["pat"][0]
    if "base" in query_params:
      self.set_base(query_params["base"][0])
    if "ui" in query_params:
      self.set_ui(query_params["ui"][0])
    if "root_certificates_path" in query_params:
      self._root_certificates_path = query_params["root_certificates_path"][0]

  @classmethod
  def from_env(cls, validate: bool = True) -> "ClarifaiAuthHelper":
    """Will look for the following env vars:
        user_id: CLARIFAI_USER_ID env var.
        app_id: CLARIFAI_APP_ID env var.
        one of:
          token: CLARIFAI_SESSION_TOKEN env var.
          pat: CLARIFAI_PAT env var.
        base: CLARIFAI_API_BASE env var.
        root_certificates_path: CLARIFAI_ROOT_CERTIFICATES_PATH env var.
    """
    user_id = os.environ.get("CLARIFAI_USER_ID", "")
    app_id = os.environ.get("CLARIFAI_APP_ID", "")
    token = os.environ.get(CLARIFAI_SESSION_TOKEN_ENV_VAR, "")
    pat = os.environ.get(CLARIFAI_PAT_ENV_VAR, "")
    base = os.environ.get("CLARIFAI_API_BASE", DEFAULT_BASE)
    ui = os.environ.get("CLARIFAI_UI", DEFAULT_UI)
    root_certificates_path = os.environ.get("CLARIFAI_ROOT_CERTIFICATES_PATH", None)
    return cls(user_id, app_id, pat, token, base, ui, root_certificates_path, validate)

  def get_user_app_id_proto(
      self,
      user_id: str = None,
      app_id: str = None,
  ) -> resources_pb2.UserAppIDSet:
    """Get the gRPC metadata that contains either the session token or the PAT to use.

    Args:
      user_id: optional user_id to override the default
      app_id: optional app_id to override the default.

    Returns:
      metadata: the metadata need to send with all grpc API calls in the API client.
    """
    user_id = self.user_id if user_id is None else user_id
    app_id = self.app_id if app_id is None else app_id
    return resources_pb2.UserAppIDSet(user_id=user_id, app_id=app_id)

  @property
  def metadata(self):
    """Get the gRPC metadata that contains either the session token or the PAT to use.

    Returns:
      metadata: the metadata need to send with all grpc API calls in the API client.
    """
    if self._pat != "":
      return (("authorization", "Key %s" % self._pat),)
    elif self._token != "":
      return (("x-clarifai-session-token", self._token),)
    else:
      raise Exception("'token' or 'pat' needed to be provided in the query params or env vars.")

  def get_stub(self) -> service_pb2_grpc.V2Stub:
    """Get the API gRPC stub using the right channel based on the API endpoint base.

    Returns:
      stub: The service_pb2_grpc.V2Stub stub for the API.
    """
    if self._base not in base_https_cache:
      raise Exception("Cannot determine if base %s is https" % self._base)

    https = base_https_cache[self._base]
    if https:
      channel = ClarifaiChannel.get_grpc_channel(
          base=self._base, root_certificates_path=self._root_certificates_path)
    else:
      if self._base.find(":") >= 0:
        host, port = self._base.split(":")
      else:
        host = self._base
        port = 80
      channel = ClarifaiChannel.get_insecure_grpc_channel(base=host, port=port)
    stub = service_pb2_grpc.V2Stub(channel)
    return stub

  @property
  def ui(self) -> str:
    """ Return the domain for the UI. """
    if self._ui not in ui_https_cache:
      raise Exception("Cannot determine if ui %s is https" % self._ui)
    https = ui_https_cache[self._ui]
    if https:
      if not self._ui.startswith("https://"):
        return "https://" + self._ui
      return self._ui
    if not self._ui.startswith("http://"):
      return "http://" + self._ui
    return self._ui

  def set_base(self, base: str):
    """ Set the base domain for the API. """
    self._base = https_cache(base_https_cache, base)

  def set_ui(self, ui: str):
    """ Set the domain for the UI. """
    self._ui = https_cache(ui_https_cache, ui)

  @property
  def base(self) -> str:
    """ Return the base domain for the API. """
    if self._base not in base_https_cache:
      raise Exception("Cannot determine if base %s is https" % self._base)
    https = base_https_cache[self._base]
    if https:
      if not self._base.startswith("https://"):
        return "https://" + self._base
      return self._base
    if not self._base.startswith("http://"):
      return "http://" + self._base
    return self._base

  @property
  def pat(self) -> str:
    """ Return the personal access token. """
    return self._pat

  def __str__(self):
    return "ClarifaiAuthHelper:\n- base: %s\n- user_id: %s\n- app_id: %s\n" % (
        self._base,
        self.user_id,
        self.app_id,
    )

  @classmethod
  def required_env_vars(cls):
    """ Return the list of the required environment variables. """
    return ["CLARIFAI_USER_ID", "CLARIFAI_APP_ID", "CLARIFAI_PAT"]

  @classmethod
  def validate_secrets_dict(cls, toml_dict: Dict[str, Any]):
    """ Validate the secrets.toml file has been filled with non-empty values for all the auth
    parameters that are present.

    We don't load the file here so that we don't need the tomli package dependency. You can simply
    do:
    import tomli

    d = tomli.load(open("secrets.toml"))
    ClarifaiAuthHelper.validate_secrets_dict(d)

    """
    # We don't validate the bases because they have sensible defaults
    auth_keys = cls.required_env_vars()

    for k, v in toml_dict.items():
      if k in auth_keys:
        if v == "":
          raise Exception("'%s' in secrets.toml cannot be empty" % k)
    # for all the keys that are not present, they have a non empty value.
    return True



import os
import time
from typing import Dict, Generator, List

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.resources_pb2 import Input
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client.base import BaseClient
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.constants.workflow import MAX_WORKFLOW_PREDICT_INPUTS
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import logger
from clarifai.utils.misc import BackoffIterator
from clarifai.workflows.export import Exporter


class Workflow(Lister, BaseClient):
  """Workflow is a class that provides access to Clarifai API endpoints related to Workflow information."""

  def __init__(self,
               url: str = None,
               workflow_id: str = None,
               workflow_version: Dict = {'id': ""},
               output_config: Dict = {'min_value': 0},
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None,
               **kwargs):
    """Initializes a Workflow object.

    Args:
        url (str): The URL to initialize the workflow object.
        workflow_id (str): The Workflow ID to interact with.
        workflow_version (dict): The Workflow Version to interact with.
        output_config (dict): The output config to interact with.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.
          sample_ms (int): The number of milliseconds to sample.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
        **kwargs: Additional keyword arguments to be passed to the Workflow.
    """
    if url and workflow_id:
      raise UserError("You can only specify one of url or workflow_id.")
    if not url and not workflow_id:
      raise UserError("You must specify one of url or workflow_id.")
    if url:
      user_id, app_id, _, workflow_id, workflow_version_id = ClarifaiUrlHelper.split_clarifai_url(
          url)
      workflow_version = {'id': workflow_version_id}
      kwargs = {'user_id': user_id, 'app_id': app_id}
    self.kwargs = {**kwargs, 'id': workflow_id, 'version': workflow_version}
    self.output_config = output_config
    self.workflow_info = resources_pb2.Workflow(**self.kwargs)
    self.logger = logger
    BaseClient.__init__(
        self,
        user_id=self.user_id,
        app_id=self.app_id,
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self)

  def predict(self, inputs: List[Input], workflow_state_id: str = None):
    """Predicts the workflow based on the given inputs.

    Args:
        inputs (list[Input]): The inputs to predict.
        workflow_state_id (str): key for the workflow state cache that stores the workflow node predictions.
    """
    if len(inputs) > MAX_WORKFLOW_PREDICT_INPUTS:
      raise UserError(f"Too many inputs. Max is {MAX_WORKFLOW_PREDICT_INPUTS}."
                     )  # TODO Use Chunker for inputs len > 32
    request = service_pb2.PostWorkflowResultsRequest(
        user_app_id=self.user_app_id,
        workflow_id=self.id,
        version_id=self.version.id,
        inputs=inputs,
        output_config=self.output_config)

    if workflow_state_id:
      request.workflow_state.id = workflow_state_id

    start_time = time.time()
    backoff_iterator = BackoffIterator(10)

    while True:
      response = self._grpc_request(self.STUB.PostWorkflowResults, request)

      if response.status.code == status_code_pb2.MODEL_DEPLOYING and \
          time.time() - start_time < 60*10:  # 10 minutes
        self.logger.info(f"{self.id} Workflow is still deploying, please wait...")
        time.sleep(next(backoff_iterator))
        continue

      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Workflow Predict failed with response {response.status!r}")
      else:
        break

    return response

  def predict_by_filepath(self, filepath: str, input_type: str):
    """Predicts the workflow based on the given filepath.

    Args:
        filepath (str): The filepath to predict.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.

    Example:
        >>> from clarifai.client.workflow import Workflow
        >>> workflow = Workflow("url") # Example: https://clarifai.com/clarifai/main/workflows/Face-Sentiment
                      or
        >>> workflow = Workflow(user_id='user_id', app_id='app_id', workflow_id='workflow_id')
        >>> workflow_prediction = workflow.predict_by_filepath('filepath', 'image')
    """
    if input_type not in {'image', 'text', 'video', 'audio'}:
      raise UserError('Invalid input type it should be image, text, video or audio.')
    if not os.path.isfile(filepath):
      raise UserError('Invalid filepath.')

    with open(filepath, "rb") as f:
      file_bytes = f.read()

    return self.predict_by_bytes(file_bytes, input_type)

  def predict_by_bytes(self, input_bytes: bytes, input_type: str):
    """Predicts the workflow based on the given bytes.

    Args:
        input_bytes (bytes): Bytes to predict on.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.
    """
    if input_type not in {'image', 'text', 'video', 'audio'}:
      raise UserError('Invalid input type it should be image, text, video or audio.')
    if not isinstance(input_bytes, bytes):
      raise UserError('Invalid bytes.')

    if input_type == "image":
      input_proto = Inputs.get_input_from_bytes("", image_bytes=input_bytes)
    elif input_type == "text":
      input_proto = Inputs.get_input_from_bytes("", text_bytes=input_bytes)
    elif input_type == "video":
      input_proto = Inputs.get_input_from_bytes("", video_bytes=input_bytes)
    elif input_type == "audio":
      input_proto = Inputs.get_input_from_bytes("", audio_bytes=input_bytes)

    return self.predict(inputs=[input_proto])

  def predict_by_url(self, url: str, input_type: str):
    """Predicts the workflow based on the given URL.

    Args:
        url (str): The URL to predict.
        input_type (str): The type of input. Can be 'image', 'text', 'video' or 'audio.

    Example:
        >>> from clarifai.client.workflow import Workflow
        >>> workflow = Workflow("url") # Example: https://clarifai.com/clarifai/main/workflows/Face-Sentiment
                      or
        >>> workflow = Workflow(user_id='user_id', app_id='app_id', workflow_id='workflow_id')
        >>> workflow_prediction = workflow.predict_by_url('url', 'image')
    """
    if input_type not in {'image', 'text', 'video', 'audio'}:
      raise UserError('Invalid input type it should be image, text, video or audio.')

    if input_type == "image":
      input_proto = Inputs.get_input_from_url("", image_url=url)
    elif input_type == "text":
      input_proto = Inputs.get_input_from_url("", text_url=url)
    elif input_type == "video":
      input_proto = Inputs.get_input_from_url("", video_url=url)
    elif input_type == "audio":
      input_proto = Inputs.get_input_from_url("", audio_url=url)

    return self.predict(inputs=[input_proto])

  def list_versions(self, page_no: int = None,
                    per_page: int = None) -> Generator['Workflow', None, None]:
    """Lists all the versions of the workflow.

    Args:
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Workflow: Workflow objects for versions of the workflow.

    Example:
        >>> from clarifai.client.workflow import Workflow
        >>> workflow = Workflow(user_id='user_id', app_id='app_id', workflow_id='workflow_id')
        >>> workflow_versions = list(workflow.list_versions())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(
        user_app_id=self.user_app_id,
        workflow_id=self.id,
    )
    all_workflow_versions_info = self.list_pages_generator(
        self.STUB.ListWorkflowVersions,
        service_pb2.ListWorkflowVersionsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for workflow_version_info in all_workflow_versions_info:
      workflow_version_info['id'] = workflow_version_info['workflow_version_id']
      del workflow_version_info['workflow_version_id']
      yield Workflow.from_auth_helper(
          auth=self.auth_helper,
          workflow_id=self.id,
          **dict(self.kwargs, version=workflow_version_info))

  def export(self, out_path: str):
    """Exports the workflow to a yaml file.

    Args:
        out_path (str): The path to save the yaml file to.

    Example:
        >>> from clarifai.client.workflow import Workflow
        >>> workflow = Workflow("https://clarifai.com/clarifai/main/workflows/Demographics")
        >>> workflow.export('out_path.yml')
    """
    request = service_pb2.GetWorkflowRequest(user_app_id=self.user_app_id, workflow_id=self.id)
    response = self._grpc_request(self.STUB.GetWorkflow, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(f"Workflow Export failed with response {response.status!r}")

    with Exporter(response) as e:
      e.parse()
      e.export(out_path)

    self.logger.info(f"Exported workflow to {out_path}")

  def __getattr__(self, name):
    return getattr(self.workflow_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.workflow_info, param)}" for param in init_params
        if hasattr(self.workflow_info, param)
    ]
    return f"Workflow Details: \n{', '.join(attribute_strings)}\n"



import logging
import os
import time
import uuid

import pytest
import yaml

from clarifai.client.user import User

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]
NOW = uuid.uuid4().hex[:10]
CREATE_APP_ID = f"ci_input_app_{NOW}"
CREATE_DATASET_ID = "ci_input_test_dataset"
CREATE_MODEL_ID = "ci_input_test_model"
CREATE_MODEL_ID_1 = "ci_input_test_model_1"
CSV_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.csv"


def create_app():
  client = User(user_id=CREATE_APP_USER_ID)
  return client.create_app(app_id=CREATE_APP_ID, base_workflow="Empty")


@pytest.mark.requires_secrets
class Testmodeltrain:
  """Tests for model training.
  """

  @classmethod
  def setup_class(self):
    self.app = create_app()
    self.input_object = self.app.inputs()
    self.dataset = self.app.create_dataset(dataset_id=CREATE_DATASET_ID)
    self.visual_classifier_model = self.app.create_model(
        model_id=CREATE_MODEL_ID, model_type_id='visual-classifier')
    self.text_classifier_model = self.app.create_model(
        model_id=CREATE_MODEL_ID_1, model_type_id='text-classifier')

  def test_model_templates(self):
    model_types = self.app.list_trainable_model_types()
    templates = self.visual_classifier_model.list_training_templates()
    assert self.visual_classifier_model.model_type_id == 'visual-classifier'  #create model test
    assert len(model_types) == 8  #list trainable model types test
    assert len(templates) >= 11  #list training templates test

  def test_model_params(self):
    model_params = self.visual_classifier_model.get_params(
        template='MMClassification_AdvancedConfig', save_to='tests/assets/model_params.yaml')
    with open('tests/assets/model_params.yaml', 'r') as file:
      expected_data = yaml.safe_load(file)
    with open(expected_data['train_params']['custom_config'], 'r') as python_file:
      custom_config = python_file.read()

    assert expected_data['train_params'].keys() == model_params[
        'train_params'].keys()  #test yaml model params
    assert len(model_params['concepts']) == 0  #test model params concepts
    assert 'custom_config' in expected_data['train_params'].keys()  #test custom config in yaml
    assert expected_data['train_params']['template'] == 'MMClassification_AdvancedConfig'  #test template param in yaml
    assert 'image_size' in expected_data['train_params'].keys()  #test image size param in yaml
    assert 'dataset_id' in expected_data.keys()  #test dataset id param in yaml
    assert '_base_' in custom_config  #test custom config script
    assert 'max_epochs' in custom_config  #test custom config script
    # cleanup
    os.remove('tests/assets/model_params.yaml')
    os.remove(expected_data['train_params']['custom_config'])

  def test_model_train(self, caplog):
    self.dataset.upload_from_csv(
        csv_path=CSV_FILE_PATH, input_type='text', csv_type='raw', labels=True)
    dataset_version = self.dataset.create_version()
    concepts = [concept.id for concept in self.app.list_concepts()]
    self.text_classifier_model.get_params(
        template='HF_GPTNeo_125m_lora', save_to='tests/assets/model_params.yaml')
    param_info = self.text_classifier_model.get_param_info(param='tokenizer_config')
    assert param_info['param'] == 'tokenizer_config'  #test get param info
    assert len(concepts) == 2  #test data upload for training
    self.text_classifier_model.update_params(
        dataset_id=CREATE_DATASET_ID,
        concepts=concepts,
        dataset_version_id=dataset_version.version.id)
    with caplog.at_level(logging.INFO):
      model_version_id = self.text_classifier_model.train()
      assert "Model Training Started" in caplog.text  #test model training

    training_status = self.text_classifier_model.training_status(version_id=model_version_id)
    assert model_version_id  #test model version id from model.train()
    assert training_status.code  #test training_status
    # NOTE (EAGLE-4139) - Immediately deleting the app causes the training to fail because the first
    # step of training requires an app, this triggers infra alerts for non graceful exit. We need to delay the deletion
    # until we get past this step.

    time.sleep(5)

    with caplog.at_level(logging.INFO):
      self.text_classifier_model.delete_version(version_id=model_version_id)
      assert "Model Version Deleted" in caplog.text  #test model version deletion
    # cleanup
    os.remove('tests/assets/model_params.yaml')

  @classmethod
  def teardown_class(self):
    self.app.delete_dataset(dataset_id=CREATE_DATASET_ID)
    self.app.delete_model(model_id=CREATE_MODEL_ID)
    self.app.delete_model(model_id=CREATE_MODEL_ID_1)
    User(user_id=CREATE_APP_USER_ID).delete_app(app_id=CREATE_APP_ID)



import time
from collections import OrderedDict, defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Tuple, Union

import numpy as np
from clarifai_grpc.grpc.api import resources_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict

from clarifai.client.dataset import Dataset
from clarifai.client.model import Model

try:
  import pandas as pd
except ImportError:
  raise ImportError("Can not import pandas. Please run `pip install pandas` to install it")

try:
  from loguru import logger
except ImportError:
  from ..logging import logger

MACRO_AVG = "macro_avg"


class EvalType(Enum):
  UNDEFINED = 0
  CLASSIFICATION = 1
  DETECTION = 2
  CLUSTERING = 3
  SEGMENTATION = 4
  TRACKER = 5


def get_eval_type(model_type):
  if "classifier" in model_type:
    return EvalType.CLASSIFICATION
  elif "visual-detector" in model_type:
    return EvalType.DETECTION
  elif "segmenter" in model_type:
    return EvalType.SEGMENTATION
  elif "embedder" in model_type:
    return EvalType.CLUSTERING
  elif "tracker" in model_type:
    return EvalType.TRACKER
  else:
    return EvalType.UNDEFINED


def to_file_name(x) -> str:
  return x.replace('/', '--')


@dataclass
class _BaseEvalResultHandler:
  model: Model
  eval_data: List[resources_pb2.EvalMetrics] = field(default_factory=list)

  def evaluate_and_wait(self, dataset: Dataset):
    from tqdm import tqdm
    dataset_id = dataset.id
    dataset_app_id = dataset.app_id
    dataset_user_id = dataset.user_id
    _ = self.model.evaluate(
        dataset_id=dataset_id, dataset_app_id=dataset_app_id, dataset_user_id=dataset_user_id)
    latest_eval = self.model.list_evaluations()[0]
    excepted = 10
    desc = f"Please wait for the evaluation process between model {self.get_model_name()} and dataset {dataset_user_id}/{dataset_app_id}/{dataset_id} to complete."
    bar = tqdm(total=excepted, desc=desc, leave=False, ncols=0)
    while latest_eval.status.code in [
        status_code_pb2.MODEL_EVALUATING, status_code_pb2.MODEL_QUEUED_FOR_EVALUATION
    ]:
      latest_eval = self.model.list_evaluations()[0]
      time.sleep(1)
      bar.update(1)

    if latest_eval.status.code == status_code_pb2.MODEL_EVALUATED:
      return latest_eval
    else:
      raise Exception(
          f"Model has failed to evaluate \n {latest_eval.status}.\nPlease check your dataset inputs!"
      )

  def find_eval_id(self, datasets: List[Dataset] = [], attempt_evaluate: bool = False):
    list_eval_outputs = self.model.list_evaluations()
    self.eval_data = []
    for dataset in datasets:
      dataset.app_id = dataset.app_id or self.model.auth_helper.app_id
      dataset.user_id = dataset.user_id or self.model.auth_helper.user_id
      dataset_assert_msg = dataset.dataset_info
      # checking if dataset exists
      out = dataset.list_versions()
      try:
        next(iter(out))
      except Exception as e:
        if any(["CONN_DOES_NOT_EXIST" in _e for _e in e.args]):
          raise Exception(
              f"Dataset {dataset_assert_msg} does not exists. Please check datasets args")
        else:
          # caused by sdk failure
          pass
      # checking if model is evaluated with this dataset
      _is_found = False
      for each in list_eval_outputs:
        if each.status.code == status_code_pb2.MODEL_EVALUATED:
          eval_dataset = each.ground_truth_dataset
          # if version_id is empty -> get latest eval result of dataset,app,user id
          if dataset.app_id == eval_dataset.app_id and dataset.id == eval_dataset.id and dataset.user_id == eval_dataset.user_id and (
              not dataset.version.id or dataset.version.id == eval_dataset.version.id):
            # append to eval_data
            self.eval_data.append(each)
            _is_found = True
            break

      # if not evaluated, but user wants to proceed it
      if not _is_found:
        if attempt_evaluate:
          self.eval_data.append(self.evaluate_and_wait(dataset))
        # otherwise raise error
        else:
          raise Exception(
              f"Model {self.model.model_info.name} in app {self.model.model_info.app_id} is not evaluated yet with dataset {dataset_assert_msg}"
          )

  @staticmethod
  def proto_to_dict(value):
    return MessageToDict(value, preserving_proto_field_name=True)

  @staticmethod
  def _f1(x: float, y: float):
    z = x + y
    return 2 * x * y / z if z else 0.

  def _get_eval(self, index=0, **kwargs):
    logger.info(
        f"Model {self.get_model_name(pretify=True)}: retrieving {kwargs} metrics of dataset: {self.get_dataset_name_by_index(index)}"
    )
    result = self.model.get_eval_by_id(eval_id=self.eval_data[index].id, **kwargs)
    for k, v in kwargs.items():
      if v:
        getattr(self.eval_data[index], k).MergeFrom(getattr(result, k))

  def get_eval_data(self, metric_name: str, index=0):
    if metric_name == 'binary_metrics':
      if len(self.eval_data[index].binary_metrics) == 0:
        self._get_eval(index, binary_metrics=True)
    elif metric_name == 'label_counts':
      if self.proto_to_dict(self.eval_data[index].label_counts) == {}:
        self._get_eval(index, label_counts=True)
    elif metric_name == 'confusion_matrix':
      if self.eval_data[index].confusion_matrix.ByteSize() == 0:
        self._get_eval(index, confusion_matrix=True)
    elif metric_name == 'metrics_by_class':
      if len(self.eval_data[index].metrics_by_class) == 0:
        self._get_eval(index, metrics_by_class=True)
    elif metric_name == 'metrics_by_area':
      if len(self.eval_data[index].metrics_by_area) == 0:
        self._get_eval(index, metrics_by_area=True)

    return getattr(self.eval_data[index], metric_name)

  def get_threshold_index(self, threshold_list: list, selected_value: float = 0.5) -> int:
    assert 0 <= selected_value <= 1 and isinstance(selected_value, float)
    threshold_list = [round(each, 2) for each in threshold_list]

    def parse_precision(x):
      return len(str(x).split(".")[1])

    precision = parse_precision(selected_value)
    if precision > 2:
      selected_value = round(selected_value, 2)
      logger.warning("Round the selected value to .2 decimals")
    return threshold_list.index(selected_value)

  def get_dataset_name_by_index(self, index=0, pretify=True):
    out = self.eval_data[index].ground_truth_dataset
    if pretify:
      app_id = out.app_id
      dataset = out.id
      #out = f"{app_id}/{dataset}/{ver[:5]}" if ver else f"{app_id}/{dataset}"
      if self.model.model_info.app_id == app_id:
        out = dataset
      else:
        out = f"{app_id}/{dataset}"

    return out

  def get_model_name(self, pretify=True):
    model = self.model.model_info
    if pretify:
      app_id = model.app_id
      name = model.id
      ver = model.model_version.id
      model = f"{app_id}/{name}/{ver[:5]}" if ver else f"{app_id}/{name}"

    return model

  def _process_curve(self, data: resources_pb2.BinaryMetrics, metric_name: str, x: str,
                     y: str) -> Dict[str, Dict[str, np.array]]:
    """ Postprocess curve
    """
    x_arr = []
    y_arr = []
    threshold = []
    outputs = []

    def _make_df(xcol, ycol, concept_col, th_col):
      return pd.DataFrame({x: xcol, y: ycol, 'concept': concept_col, 'threshold': th_col})

    for bd in data:
      concept_id = bd.concept.id
      metric = eval(f'bd.{metric_name}')
      if metric.ByteSize() == 0:
        continue
      _x = np.array(eval(f'metric.{x}'))
      _y = np.array(eval(f'metric.{y}'))
      threshold = np.array(metric.thresholds)
      x_arr.append(_x)
      y_arr.append(_y)
      concept_cols = [concept_id for _ in range(len(_x))]
      outputs.append(_make_df(_x, _y, concept_cols, threshold))

    avg_x = np.mean(x_arr, axis=0)
    avg_y = np.mean(y_arr, axis=0)
    if np.isnan(avg_x).all():
      return None
    else:
      avg_cols = [MACRO_AVG for _ in range(len(avg_x))]
    outputs.append(_make_df(avg_x, avg_y, avg_cols, threshold))

    return pd.concat(outputs, axis=0)

  def parse_concept_ids(self, *args, **kwargs) -> List[str]:
    raise NotImplementedError

  def detailed_summary(self, *args, **kwargs):
    raise NotImplementedError

  def pr_curve(self, *args, **kwargs):
    raise NotImplementedError

  def roc_curve(self, *args, **kwargs):
    raise NotImplementedError

  def confusion_matrix(self, *args, **kwargs):
    raise NotImplementedError


@dataclass
class PlaceholderHandler(_BaseEvalResultHandler):

  def parse_concept_ids(self, *args, **kwargs) -> List[str]:
    return None

  def detailed_summary(self, *args, **kwargs):
    return None

  def pr_curve(self, *args, **kwargs):
    return None


@dataclass
class ClassificationResultHandler(_BaseEvalResultHandler):

  def parse_concept_ids(self, index=0) -> List[str]:
    eval_data = self.get_eval_data(metric_name='label_counts', index=index)
    concept_ids = [temp.concept.id for temp in eval_data.positive_label_counts]
    return concept_ids

  def detailed_summary(self, index=0, confidence_threshold: float = 0.5,
                       **kwargs) -> Union[None, Tuple[pd.DataFrame, pd.DataFrame]]:
    """Making detailed table per concept and for total concept

    Args:
        index (int, optional): Index of eval dataset. Defaults to 0.
        confidence_threshold (float, optional): confidence threshold. Defaults to 0.5.

    Returns:
        tuple: concepts dataframe, total dataframe
    """
    eval_data = self.get_eval_data('binary_metrics', index=index)
    summary = self.get_eval_data('summary', index=index)

    total_labeled = 0
    total_predicted = 0
    total_tp = 0
    total_fn = 0
    total_fp = 0
    metrics = []

    for bd in eval_data:
      concept_id = bd.concept.id
      if bd.precision_recall_curve.ByteSize() == 0:
        continue
      pr_th_index = self.get_threshold_index(
          list(bd.precision_recall_curve.thresholds), selected_value=confidence_threshold)
      roc_th_index = self.get_threshold_index(
          list(bd.roc_curve.thresholds), selected_value=confidence_threshold)
      if pr_th_index is None or roc_th_index is None:
        continue
      num_pos_labeled = bd.num_pos
      num_neg_labeled = bd.num_neg
      # TP/(TP+FP)
      precision = bd.precision_recall_curve.precision[pr_th_index]
      # TP/(TP+FN)
      recall = bd.precision_recall_curve.recall[pr_th_index]
      # FP/(FP+TN)
      fpr = bd.roc_curve.fpr[roc_th_index]
      # TP/(TP+FN)
      tpr = bd.roc_curve.tpr[roc_th_index]
      # TP+FN
      tp = int(tpr * num_pos_labeled)
      fn = num_pos_labeled - tp
      fp = int(fpr * num_neg_labeled)
      num_pos_pred = tp + fp
      f1 = self._f1(recall, precision)

      total_labeled += num_pos_labeled
      total_predicted += num_pos_pred
      total_fn += fn
      total_tp += tp
      total_fp += fp
      # roc auc, total labelled, predicted, tp, fn, fp, recall, precision, f1
      _d = OrderedDict({
          "Concept": concept_id,
          "Accuracy (ROC AUC)": round(bd.roc_auc, 3),
          "Total Labeled": num_pos_labeled,
          "Total Predicted": num_pos_pred,
          "True Positives": tp,
          "False Negatives": fn,
          "False Positives": fp,
          "Recall": recall,
          "Precision": precision,
          "F1": f1
      })
      metrics.append(pd.DataFrame(_d, index=[0]))

    # If no valid data is found, return None
    if not metrics:
      return None
    # Make per concept df
    df = pd.concat(metrics, axis=0)
    # Make total df
    sum_df_total = sum(df["Total Labeled"])
    precision = sum(df.Precision * df["Total Labeled"]) / sum_df_total if sum_df_total else 0.
    recall = sum(df.Recall * df["Total Labeled"]) / sum_df_total if sum_df_total else 0.
    f1 = self._f1(recall, precision)
    df_total = pd.DataFrame(
        [
            [
                'Total', summary.macro_avg_roc_auc, total_labeled, total_predicted, total_tp,
                total_fn, total_fp, recall, precision, f1
            ],
        ],
        columns=df.columns,
        index=[0])

    return df, df_total

  def pr_curve(self, index=0, **kwargs) -> Union[None, pd.DataFrame]:
    """Making PR curve

    Args:
        index (int, optional): Index of eval dataset. Defaults to 0.

    Returns:
        dictionary: Keys are concept ids and 'macro_avg'. Values are dictionaries of {precision: np.array, recall: np.array}
    """
    eval_data = self.get_eval_data(metric_name='binary_metrics', index=index)
    outputs = self._process_curve(
        eval_data, metric_name='precision_recall_curve', x='recall', y='precision')
    return outputs

  def roc_curve(self, index=0, **kwargs) -> Union[None, pd.DataFrame]:
    eval_data = self.get_eval_data(metric_name='binary_metrics', index=index)
    outputs = self._process_curve(eval_data, metric_name='roc_curve', x='tpr', y='fpr')
    return outputs

  def confusion_matrix(self, index=0, **kwargs):
    eval_data = self.get_eval_data(metric_name='confusion_matrix', index=index)
    concept_ids = self.parse_concept_ids(index)
    concept_ids.sort()
    data = np.zeros((len(concept_ids), len(concept_ids)), np.float32)
    for entry in eval_data.matrix:
      p = entry.predicted_concept.id
      a = entry.actual_concept.id
      if p in concept_ids and a in concept_ids:
        data[concept_ids.index(a), concept_ids.index(p)] = np.around(entry.value, decimals=3)
      else:
        continue
    rownames = pd.MultiIndex.from_arrays([concept_ids], names=['Actual'])
    colnames = pd.MultiIndex.from_arrays([concept_ids], names=['Predicted'])
    df = pd.DataFrame(data, columns=colnames, index=rownames)

    return df


@dataclass
class DetectionResultHandler(_BaseEvalResultHandler):
  AREA_LIST = ["all", "medium", "small"]
  IOU_LIST = list(np.arange(0.5, 1., 0.1))

  def parse_concept_ids(self, index=0) -> List[str]:
    eval_data = self.get_eval_data(metric_name='metrics_by_class', index=index)
    concept_ids = [temp.concept.id for temp in eval_data]
    return concept_ids

  def detailed_summary(self,
                       index=0,
                       confidence_threshold: float = 0.5,
                       iou_threshold: float = 0.5,
                       area: str = "all",
                       bypass_const: bool = False,
                       **kwargs):
    if not bypass_const:
      assert iou_threshold in self.IOU_LIST, f"Expected iou_threshold in {self.IOU_LIST}, got {iou_threshold}"
      assert area in self.AREA_LIST, f"Expected area in {self.AREA_LIST}, got {area}"

    eval_data = self.get_eval_data('metrics_by_class', index=index)
    #summary = self.get_eval_data('summary', index=index)
    metrics = []
    for bd in eval_data:
      # total label
      _iou = round(bd.iou, 1)
      if not (area and bd.area_name == area) or not (iou_threshold and iou_threshold == _iou):
        continue
      concept_id = bd.concept.id
      total = round(bd.num_tot, 3)
      # TP / (TP + FP)
      if len(bd.precision_recall_curve.precision) > 0:
        pr_th_index = self.get_threshold_index(
            list(bd.precision_recall_curve.thresholds), selected_value=confidence_threshold)
        p = round(bd.precision_recall_curve.precision[pr_th_index], 3)
      else:
        p = 0
      # TP / (TP + FN)
      if len(bd.precision_recall_curve.recall) > 0:
        pr_th_index = self.get_threshold_index(
            list(bd.precision_recall_curve.thresholds), selected_value=confidence_threshold)
        r = round(bd.precision_recall_curve.recall[pr_th_index], 3)
      else:
        r = 0
      tp = int(round(r * total, 0))
      fn = total - tp
      fp = float(tp) / p - tp if p else 0
      fp = int(round(fp, 1))
      f1 = self._f1(r, p)
      _d = {
          "Concept": concept_id,
          "Average Precision": round(float(bd.avg_precision), 3),
          "Total Labeled": total,
          "True Positives": tp,
          "False Positives": fp,
          "False Negatives": fn,
          "Recall": r,
          "Precision": p,
          "F1": f1,
      }
      metrics.append(pd.DataFrame(_d, index=[0]))

    if not metrics:
      return None

    df = pd.concat(metrics, axis=0)
    df_total = defaultdict()
    sum_df_total = df["Total Labeled"].sum()
    df_total["Concept"] = "Total"
    df_total["Average Precision"] = df["Average Precision"].mean()
    df_total["Total Labeled"] = sum_df_total
    df_total["True Positives"] = df["True Positives"].sum()
    df_total["False Positives"] = df["False Positives"].sum()
    df_total["False Negatives"] = df["False Negatives"].sum()
    df_total["Recall"] = sum(
        df.Recall * df["Total Labeled"]) / sum_df_total if sum_df_total else 0.
    df_total["Precision"] = df_total["True Positives"] / (
        df_total["True Positives"] + df_total["False Positives"]) if sum_df_total else 0.
    df_total["F1"] = self._f1(df_total["Recall"], df_total["Precision"])
    df_total = pd.DataFrame(df_total, index=[0])

    return [df, df_total]

  def pr_curve(self,
               index=0,
               iou_threshold: float = 0.5,
               area: str = "all",
               bypass_const=False,
               **kwargs):

    if not bypass_const:
      assert iou_threshold in self.IOU_LIST, f"Expected iou_threshold in {self.IOU_LIST}, got {iou_threshold}"
      assert area in self.AREA_LIST, f"Expected area in {self.AREA_LIST}, got {area}"

    eval_data = self.get_eval_data(metric_name='metrics_by_class', index=index)
    _valid_eval_data = []
    for bd in eval_data:
      _iou = round(bd.iou, 1)
      if not (area and bd.area_name == area) or not (iou_threshold and iou_threshold == _iou):
        continue
      _valid_eval_data.append(bd)

    outputs = self._process_curve(
        _valid_eval_data, metric_name='precision_recall_curve', x='recall', y='precision')
    return outputs

  def roc_curve(self, index=0, **kwargs) -> None:
    return None

  def confusion_matrix(self, index=0, **kwargs) -> None:
    return None


def make_handler_by_type(model_type: str) -> _BaseEvalResultHandler:
  _eval_type = get_eval_type(model_type)
  if _eval_type == EvalType.CLASSIFICATION:
    return ClassificationResultHandler
  elif _eval_type == EvalType.DETECTION:
    return DetectionResultHandler
  else:
    return PlaceholderHandler



import logging
import os
import uuid

import pytest

from clarifai.client.user import User

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]
NOW = uuid.uuid4().hex[:10]
CREATE_APP_ID = f"ci_input_app_{NOW}"
CREATE_DATASET_ID = "ci_input_test_dataset"
CREATE_MODEL_ID = "ci_input_test_model_1"
CSV_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.csv"


def create_app():
  client = User(user_id=CREATE_APP_USER_ID)
  return client.create_app(app_id=CREATE_APP_ID, base_workflow="Empty")


@pytest.mark.requires_secrets
class TestEval:
  """Tests for model evaluating.
  """

  @classmethod
  def setup_class(self):
    self.app = create_app()
    self.input_object = self.app.inputs()
    self.dataset = self.app.create_dataset(dataset_id=CREATE_DATASET_ID)
    self.model = self.app.create_model(model_id=CREATE_MODEL_ID, model_type_id='text-classifier')

  def test_evaluate(self, caplog):
    # Prepare dataset
    self.dataset.upload_from_csv(
        csv_path=CSV_FILE_PATH, input_type='text', csv_type='raw', labels=True)
    dataset_version = self.dataset.create_version()
    concepts = [concept.id for concept in self.app.list_concepts()]
    # Prepare for training
    self.model.get_params(
        template='HF_GPTNeo_125m_lora', save_to='tests/assets/model_params_for_eval.yaml')
    param_info = self.model.get_param_info(param='tokenizer_config')
    assert param_info['param'] == 'tokenizer_config'  #test get param info
    assert len(concepts) == 2  #test data upload for training
    self.model.update_params(
        dataset_id=CREATE_DATASET_ID,
        concepts=concepts,
        dataset_version_id=dataset_version.version.id)
    with caplog.at_level(logging.INFO):
      model_version_id = self.model.train()
      assert "Model Training Started" in caplog.text  #test model training

    # Test eval
    ## Test list eval
    with pytest.raises(Exception) as e:
      self.model.list_evaluations()
    assert "model version is empty" in str(e.value).lower()

    self.model.model_info.model_version.id = model_version_id
    with pytest.raises(Exception) as e:
      self.model.list_evaluations()
    assert "model not yet trained" in str(e.value).lower()
    ## Test evaluate
    with pytest.raises(Exception) as e:
      self.model.evaluate(dataset_id=CREATE_DATASET_ID)
    assert "model not yet trained" in str(e.value).lower()

    ## Test get eval
    with pytest.raises(Exception) as e:
      self.model.get_latest_eval(
          label_counts=True,
          test_set=True,
          binary_metrics=True,
          confusion_matrix=True,
          metrics_by_area=True,
          metrics_by_class=True)
    assert "model not yet trained" in str(e.value).lower()

    # cleanup
    with caplog.at_level(logging.INFO):
      self.model.delete_version(version_id=model_version_id)
      assert "Model Version Deleted" in caplog.text  #test model version deletion
    os.remove('tests/assets/model_params_for_eval.yaml')

  @classmethod
  def teardown_class(self):
    self.app.delete_dataset(dataset_id=CREATE_DATASET_ID)
    self.app.delete_model(model_id=CREATE_MODEL_ID)
    User(user_id=CREATE_APP_USER_ID).delete_app(app_id=CREATE_APP_ID)



"""
This is simply the main file for the server that imports ModelRunner implementation
and starts the server.
"""

import argparse
import importlib.util
import inspect
import os
import sys
from concurrent import futures

from clarifai_grpc.grpc.api import service_pb2_grpc
from clarifai_protocol import BaseRunner
from clarifai_protocol.utils.grpc_server import GRPCServer

from clarifai.runners.models.model_servicer import ModelServicer
from clarifai.utils.logging import logger


def main():
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--port',
      type=int,
      default=8000,
      help="The port to host the gRPC server at.",
      choices=range(1024, 65535),
  )
  parser.add_argument(
      '--pool_size',
      type=int,
      default=32,
      help="The number of threads to use for the gRPC server.",
      choices=range(1, 129),
  )  # pylint: disable=range-builtin-not-iterating
  parser.add_argument(
      '--max_queue_size',
      type=int,
      default=10,
      help='Max queue size of requests before we begin to reject requests (default: 10).',
      choices=range(1, 21),
  )  # pylint: disable=range-builtin-not-iterating
  parser.add_argument(
      '--max_msg_length',
      type=int,
      default=1024 * 1024 * 1024,
      help='Max message length of grpc requests (default: 1 GB).',
  )
  parser.add_argument(
      '--enable_tls',
      action='store_true',
      default=False,
      help=
      'Set to true to enable TLS (default: False) since this server is meant for local development only.',
  )
  parser.add_argument(
      '--start_dev_server',
      action='store_true',
      default=False,
      help=
      'Set to true to start the gRPC server (default: False). If set to false, the server will not start and only the runner loop will start to fetch work from the API.',
  )
  parser.add_argument(
      '--model_path',
      type=str,
      required=True,
      help='The path to the model directory that contains implemention of the model.',
  )

  parsed_args = parser.parse_args()

  # import the runner class that to be implement by the user
  runner_path = os.path.join(parsed_args.model_path, "1", "model.py")

  # arbitrary name given to the module to be imported
  module = "runner_module"

  spec = importlib.util.spec_from_file_location(module, runner_path)
  runner_module = importlib.util.module_from_spec(spec)
  sys.modules[module] = runner_module
  spec.loader.exec_module(runner_module)

  # Find all classes in the model.py file that are subclasses of BaseRunner
  classes = [
      cls for _, cls in inspect.getmembers(runner_module, inspect.isclass)
      if issubclass(cls, BaseRunner) and cls.__module__ == runner_module.__name__
  ]

  #  Ensure there is exactly one subclass of BaseRunner in the model.py file
  if len(classes) != 1:
    raise Exception("Expected exactly one subclass of BaseRunner, found: {}".format(len(classes)))

  MyRunner = classes[0]

  # initialize the Runner class. This is what the user implements.
  # (Note) do we want to set runner_id, nodepool_id, compute_cluster_id, base_url, num_parallel_polls as env vars? or as args?
  runner = MyRunner(
      runner_id=os.environ["CLARIFAI_RUNNER_ID"],
      nodepool_id=os.environ["CLARIFAI_NODEPOOL_ID"],
      compute_cluster_id=os.environ["CLARIFAI_COMPUTE_CLUSTER_ID"],
      base_url=os.environ["CLARIFAI_API_BASE"],
      num_parallel_polls=int(os.environ.get("CLARIFAI_NUM_THREADS", 1)),
  )

  # initialize the servicer
  servicer = ModelServicer(runner)

  # Setup the grpc server for local development.
  if parsed_args.start_dev_server:
    server = GRPCServer(
        futures.ThreadPoolExecutor(
            max_workers=parsed_args.pool_size,
            thread_name_prefix="ServeCalls",
        ),
        parsed_args.max_msg_length,
        parsed_args.max_queue_size,
    )
    server.add_port_to_server('[::]:%s' % parsed_args.port, parsed_args.enable_tls)

    service_pb2_grpc.add_V2Servicer_to_server(servicer, server)
    server.start()
    logger.info("Started server on port %s", parsed_args.port)
    # server.wait_for_termination() # won't get here currently.

  runner.start()  # start the runner loop to fetch work from the API.


if __name__ == '__main__':
  main()



from typing import Any, Callable, Dict, Generator

from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict

from clarifai.client.base import BaseClient


class Lister(BaseClient):
  """Lister class for obtaining paginated results from the Clarifai API."""

  def __init__(self, page_size: int = 16):
    self.default_page_size = page_size

  def list_pages_generator(self,
                           endpoint: Callable,
                           proto_message: Any,
                           request_data: Dict[str, Any],
                           page_no: int = None,
                           per_page: int = None) -> Generator[Dict[str, Any], None, None]:
    """Lists pages of a resource.

    Args:
        endpoint (Callable): The endpoint to call.
        proto_message (Any): The proto message to use.
        request_data (dict): The request data to use.
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        response_dict: The next item in the listing.
    """
    page = 1 if not page_no else page_no
    if page_no and not per_page:
      per_page = self.default_page_size
    while True:
      request_data['page'] = page
      request_data['per_page'] = per_page
      response = self._grpc_request(endpoint, proto_message(**request_data))
      dict_response = MessageToDict(response, preserving_proto_field_name=True)
      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Listing failed with response {response!r}")
      if len(list(dict_response.keys())) == 1:
        break
      else:
        listing_resource = list(dict_response.keys())[1]
        for item in dict_response[listing_resource]:
          if listing_resource == "dataset_inputs":
            yield self.process_response_keys(item["input"], listing_resource[:-1])
          else:
            yield self.process_response_keys(item, listing_resource[:-1])
      if page_no is not None or per_page is not None:
        break
      page += 1



import os
from typing import Any, Dict, Generator, List

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict

from clarifai.client.base import BaseClient
from clarifai.client.lister import Lister
from clarifai.client.nodepool import Nodepool
from clarifai.errors import UserError
from clarifai.utils.logging import logger


class ComputeCluster(Lister, BaseClient):
  """ComputeCluster is a class that provides access to Clarifai API endpoints related to Compute Cluster information."""

  def __init__(self,
               compute_cluster_id: str = None,
               user_id: str = None,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None,
               **kwargs):
    """Initializes an ComputeCluster object.

    Args:
        compute_cluster_id (str): The ComputeCluster ID for the ComputeCluster to interact with.
        user_id (str): The user ID of the user.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
        **kwargs: Additional keyword arguments to be passed to the compute cluster.
    """
    self.kwargs = {**kwargs, 'id': compute_cluster_id, 'user_id': user_id}
    self.compute_cluster_info = resources_pb2.ComputeCluster(**self.kwargs)
    self.logger = logger
    BaseClient.__init__(
        self,
        user_id=self.user_id,
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self)

  def list_nodepools(self, page_no: int = None,
                     per_page: int = None) -> Generator[Nodepool, None, None]:
    """Lists all the available nodepools of the compute cluster.

    Args:
        compute_cluster_id (str): The compute cluster ID to list the nodepools.
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Nodepool: Nodepool objects for the nodepools in the compute cluster.

    Example:
        >>> from clarifai.client.compute_cluster import ComputeCluster
        >>> compute_cluster = ComputeCluster(compute_cluster_id="compute_cluster_id", user_id="user_id")
        >>> all_nodepools = list(compute_cluster.list_nodepools())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(user_app_id=self.user_app_id, compute_cluster_id=self.id)
    all_nodepools_info = self.list_pages_generator(
        self.STUB.ListNodepools,
        service_pb2.ListNodepoolsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for nodepool_info in all_nodepools_info:
      yield Nodepool.from_auth_helper(auth=self.auth_helper, **nodepool_info)

  def _process_nodepool_config(self, config_filepath: str) -> Dict[str, Any]:
    with open(config_filepath, "r") as file:
      nodepool_config = yaml.safe_load(file)

    assert "nodepool" in nodepool_config, "nodepool info not found in the config file"
    nodepool = nodepool_config['nodepool']
    assert "instance_types" in nodepool, "region not found in the config file"
    assert "node_capacity_type" in nodepool, "managed_by not found in the config file"
    assert "max_instances" in nodepool, "cluster_type not found in the config file"
    nodepool['compute_cluster'] = resources_pb2.ComputeCluster(id=self.id, user_id=self.user_id)
    nodepool['node_capacity_type'] = resources_pb2.NodeCapacityType(capacity_types=[
        capacity_type for capacity_type in nodepool['node_capacity_type']['capacity_types']
    ])
    instance_types = []
    for instance_type in nodepool['instance_types']:
      if 'compute_info' in instance_type:
        instance_type['compute_info'] = resources_pb2.ComputeInfo(**instance_type['compute_info'])
      instance_types.append(resources_pb2.InstanceType(**instance_type))
    nodepool['instance_types'] = instance_types
    if "visibility" in nodepool:
      nodepool["visibility"] = resources_pb2.Visibility(**nodepool["visibility"])
    return nodepool

  def create_nodepool(self, nodepool_id: str, config_filepath: str) -> Nodepool:
    """Creates a nodepool for the compute cluster.

    Args:
        nodepool_id (str): The nodepool ID for the nodepool to create.
        config_filepath (str): The path to the nodepool config file.

    Returns:
        Nodepool: A Nodepool object for the specified nodepool ID.

    Example:
        >>> from clarifai.client.compute_cluster import ComputeCluster
        >>> compute_cluster = ComputeCluster(compute_cluster_id="compute_cluster_id", user_id="user_id")
        >>> nodepool = compute_cluster.create_nodepool(nodepool_id="nodepool_id", config_filepath="config.yml")
    """
    if not os.path.exists(config_filepath):
      raise UserError(f"Nodepool config file not found at {config_filepath}")

    nodepool_config = self._process_nodepool_config(config_filepath)

    if 'id' in nodepool_config:
      nodepool_id = nodepool_config['id']
      nodepool_config.pop('id')

    request = service_pb2.PostNodepoolsRequest(
        user_app_id=self.user_app_id,
        compute_cluster_id=self.id,
        nodepools=[resources_pb2.Nodepool(id=nodepool_id, **nodepool_config)])
    response = self._grpc_request(self.STUB.PostNodepools, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nNodepool created\n%s", response.status)

    return Nodepool.from_auth_helper(self.auth_helper, nodepool_id=nodepool_id)

  def nodepool(self, nodepool_id: str) -> Nodepool:
    """Returns a Nodepool object for the existing nodepool ID.

    Args:
        nodepool_id (str): The nodepool ID for the nodepool to interact with.

    Returns:
        Nodepool: A Nodepool object for the existing nodepool ID.

    Example:
        >>> from clarifai.client.compute_cluster import ComputeCluster
        >>> compute_cluster = ComputeCluster(compute_cluster_id="compute_cluster_id", user_id="user_id")
        >>> nodepool = compute_cluster.nodepool(nodepool_id="nodepool_id")
    """
    request = service_pb2.GetNodepoolRequest(
        user_app_id=self.user_app_id, compute_cluster_id=self.id, nodepool_id=nodepool_id)
    response = self._grpc_request(self.STUB.GetNodepool, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    kwargs = self.process_response_keys(dict_response[list(dict_response.keys())[1]],
                                        list(dict_response.keys())[1])

    return Nodepool.from_auth_helper(auth=self.auth_helper, **kwargs)

  def delete_nodepools(self, nodepool_ids: List[str]) -> None:
    """Deletes list of nodepools for the compute cluster.

    Args:
        nodepool_ids (List[str]): The nodepool IDs of the compute cluster to delete.

    Example:
        >>> from clarifai.client.compute_cluster import ComputeCluster
        >>> compute_cluster = ComputeCluster(compute_cluster_id="compute_cluster_id", user_id="user_id")
        >>> compute_cluster.delete_nodepools(nodepool_ids=["nodepool_id1", "nodepool_id2"])
    """
    assert isinstance(nodepool_ids, list), "nodepool_ids param should be a list"

    request = service_pb2.DeleteNodepoolsRequest(
        user_app_id=self.user_app_id, compute_cluster_id=self.id, ids=nodepool_ids)
    response = self._grpc_request(self.STUB.DeleteNodepools, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nNodepools Deleted\n%s", response.status)

  def __getattr__(self, name):
    return getattr(self.compute_cluster_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.compute_cluster_info, param)}" for param in init_params
        if hasattr(self.compute_cluster_info, param)
    ]
    return f"Clarifai Compute Cluster Details: \n{', '.join(attribute_strings)}\n"



import io
from pathlib import Path
from typing import List

import requests


## TODO: Make this token-aware.
def convert_messages_to_str(messages: List[dict]) -> str:
  """convert messages in OpenAI API format into a single string.

    Args:
        messages List[dict]: A list of dictionary in the following format:
        ```
        [
          {"role": "user", "content": "Hello there."},
          {"role": "assistant", "content": "Hi, I'm Claude. How can I help you?"},
          {"role": "user", "content": "Can you explain LLMs in plain English?"},
        ]
        ```
    """
  final_str = ""
  for msg in messages:
    if "role" in msg and "content" in msg:
      role = msg.get("role", "")
      content = msg.get("content", "")
      final_str += f"\n\n{role}: {content}"
  return final_str


def format_assistant_message(raw_text: str) -> dict:
  return {"role": "assistant", "content": raw_text}


def load_documents(file_path: str = None, folder_path: str = None, url: str = None) -> List[any]:
  """Loads documents from a local directory or public url or local filename.

  Args:
      file_path (str): The path to the filename.
      folder_path (str): The path to the folder.
      url (str): The url to the file.
  """
  #check import packages
  try:
    from llama_index.core import Document, SimpleDirectoryReader
    from llama_index.core.readers.download import download_loader
  except ImportError:
    raise ImportError("Could not import llama index package. "
                      "Please install it with `pip install llama-index-core==0.10.1`.")
  #document loaders for filepath
  if file_path:
    if file_path.endswith(".pdf"):
      PDFReader = download_loader("PDFReader")
      loader = PDFReader()
      documents = loader.load_data(file=Path(file_path))
    elif file_path.endswith(".docx"):
      docReader = download_loader("DocxReader")
      loader = docReader()
      documents = loader.load_data(file=Path(file_path))
    elif file_path.endswith(".txt"):
      with open(file_path, 'r') as file:
        text_content = file.read()
      documents = [Document(text=text_content)]
    else:
      raise ValueError("Only .pdf, .docx, and .txt files are supported.")

  #document loaders for folderpath
  if folder_path:
    documents = SimpleDirectoryReader(
        input_dir=Path(folder_path), required_exts=[".pdf", ".docx", ".txt"]).load_data()

  #document loaders for url
  if url:
    response = requests.get(url)
    if response.status_code != 200:
      raise ValueError(f"Invalid url {url}.")
    #for text files
    try:
      documents = [Document(text=response.content)]
    #for pdf files
    except Exception:
      #check import packages
      try:
        from pypdf import PdfReader
      except ImportError:
        raise ImportError("Could not import pypdf package. "
                          "Please install it with `pip install pypdf==3.17.4`.")
      documents = []
      pdf_file = PdfReader(io.BytesIO(response.content))
      num_pages = len(pdf_file.pages)
      for page in range(num_pages):
        page_text = pdf_file.pages[page].extract_text()
        documents.append(Document(text=page_text))
    else:
      raise ValueError(f"Invalid url {url}.")

  return documents


def split_document(text: str, chunk_size: int, chunk_overlap: int, **kwargs) -> List[str]:
  """Splits a document into chunks of text.

  Args:
      text (str): The text to split.
      chunk_size (int): The size of each chunk.
      chunk_overlap (int): The amount of overlap between each chunk.
      **kwargs: Additional keyword arguments for the SentenceSplitter.
  """
  #check import packages
  try:
    from llama_index.core.node_parser.text import SentenceSplitter
  except ImportError:
    raise ImportError("Could not import llama index package. "
                      "Please install it with `pip install llama-index-core==0.10.24`.")
  #document
  text_parser = SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap, **kwargs)
  text_chunks = text_parser.split_text(text)
  return text_chunks



import importlib.util
import json
import os
import subprocess

from clarifai.utils.logging import logger


class HuggingFaceLoarder:

  def __init__(self, repo_id=None, token=None):
    self.repo_id = repo_id
    self.token = token
    if token:
      try:
        if importlib.util.find_spec("huggingface_hub") is None:
          raise ImportError(
              "The 'huggingface_hub' package is not installed. Please install it using 'pip install huggingface_hub'."
          )
        os.environ['HF_TOKEN'] = token
        subprocess.run(f'huggingface-cli login --token={os.environ["HF_TOKEN"]}', shell=True)
      except Exception as e:
        Exception("Error setting up Hugging Face token ", e)

  def download_checkpoints(self, checkpoint_path: str):
    # throw error if huggingface_hub wasn't installed
    try:
      from huggingface_hub import snapshot_download
    except ImportError:
      raise ImportError(
          "The 'huggingface_hub' package is not installed. Please install it using 'pip install huggingface_hub'."
      )
    if os.path.exists(checkpoint_path) and self.validate_download(checkpoint_path):
      logger.info("Checkpoints already exist")
      return True
    else:
      os.makedirs(checkpoint_path, exist_ok=True)
      try:
        is_hf_model_exists = self.validate_hf_model()
        if not is_hf_model_exists:
          logger.error("Model %s not found on Hugging Face" % (self.repo_id))
          return False
        snapshot_download(
            repo_id=self.repo_id, local_dir=checkpoint_path, local_dir_use_symlinks=False)
      except Exception as e:
        logger.exception(f"Error downloading model checkpoints {e}")
        return False
      finally:
        is_downloaded = self.validate_download(checkpoint_path)
        if not is_downloaded:
          logger.error("Error validating downloaded model checkpoints")
          return False
      return True

  def validate_hf_model(self,):
    # check if model exists on HF

    from huggingface_hub import file_exists, repo_exists
    return repo_exists(self.repo_id) and file_exists(self.repo_id, 'config.json')

  def validate_download(self, checkpoint_path: str):
    # check if model exists on HF
    from huggingface_hub import list_repo_files
    checkpoint_dir_files = [
        f for dp, dn, fn in os.walk(os.path.expanduser(checkpoint_path)) for f in fn
    ]
    return (len(checkpoint_dir_files) >= len(list_repo_files(self.repo_id))) and len(
        list_repo_files(self.repo_id)) > 0

  def fetch_labels(self, checkpoint_path: str):
    # Fetch labels for classification, detection and segmentation models
    config_path = os.path.join(checkpoint_path, 'config.json')
    with open(config_path, 'r') as f:
      config = json.load(f)

    labels = config['id2label']
    return labels



#! dataset output features (output from preprocessing & input to clarifai data proto builders)
from dataclasses import dataclass
from typing import List, Optional, Union


@dataclass
class TextFeatures:
  """Text classification datasets preprocessing output features."""
  text: str
  labels: List[Union[str, int]]  # List[str or int] to cater for multi-class tasks
  id: Optional[int] = None  # text_id
  metadata: Optional[dict] = None
  label_ids: Optional[List[str]] = None


@dataclass
class VisualClassificationFeatures:
  """Image classification datasets preprocessing output features."""
  image_path: str
  labels: List[Union[str, int]]  # List[str or int] to cater for multi-class tasks
  geo_info: Optional[List[float]] = None  #[Longitude, Latitude]
  id: Optional[int] = None  # image_id
  metadata: Optional[dict] = None
  image_bytes: Optional[bytes] = None
  label_ids: Optional[List[str]] = None


@dataclass
class VisualDetectionFeatures:
  """Image Detection datasets preprocessing output features."""
  image_path: str
  labels: List[Union[str, int]]
  bboxes: List[List[float]]
  geo_info: Optional[List[float]] = None  #[Longitude, Latitude]
  id: Optional[int] = None  # image_id
  metadata: Optional[dict] = None
  image_bytes: Optional[bytes] = None
  label_ids: Optional[List[str]] = None


@dataclass
class VisualSegmentationFeatures:
  """Image Segmentation datasets preprocessing output features."""
  image_path: str
  labels: List[Union[str, int]]
  polygons: List[List[List[float]]]
  geo_info: Optional[List[float]] = None  #[Longitude, Latitude]
  id: Optional[int] = None  # image_id
  metadata: Optional[dict] = None
  image_bytes: Optional[bytes] = None
  label_ids: Optional[List[str]] = None


@dataclass
class MultiModalFeatures:
  """Multi-modal datasets preprocessing output features."""
  text: str
  image_bytes: str
  labels: List[Union[str, int]] = None  # List[str or int] to cater for multi-class tasks
  id: Optional[int] = None  # image_id
  metadata: Optional[dict] = None



import abc
import grpc

from clarifai_grpc.grpc.api import service_pb2_grpc


class V2Stub(abc.ABC):
  """Abstract base class of clarifai api rpc client stubs"""


class RpcCallable(abc.ABC):
  """Abstract base class of clarifai api rpc callables"""


# add grpc classes as subclasses of the abcs, so they also succeed in isinstance calls
def _register_classes():
  V2Stub.register(service_pb2_grpc.V2Stub)
  for name in dir(grpc):
    if name.endswith('Callable'):
      RpcCallable.register(getattr(grpc, name))


_register_classes()



import uuid
from typing import List

import yaml
from clarifai_grpc.grpc.api import resources_pb2  # noqa: F401
from google.protobuf.struct_pb2 import Struct

from clarifai.client.app import App
from clarifai.client.input import Inputs
from clarifai.client.model import Model
from clarifai.client.user import User
from clarifai.client.workflow import Workflow
from clarifai.constants.rag import MAX_UPLOAD_BATCH_SIZE
from clarifai.errors import UserError
from clarifai.rag.utils import (convert_messages_to_str, format_assistant_message, load_documents,
                                split_document)
from clarifai.utils.constants import CLARIFAI_USER_ID_ENV_VAR
from clarifai.utils.logging import logger
from clarifai.utils.misc import get_from_dict_or_env

DEFAULT_RAG_PROMPT_TEMPLATE = "Context information is below:\n{data.hits}\nGiven the context information and not prior knowledge, answer the query.\nQuery: {data.text.raw}\nAnswer: "


class RAG:
  """
    RAG is a class for Retrieval Augmented Generation.

    Example:
        >>> from clarifai.rag import RAG
        >>> rag_agent = RAG(workflow_url=YOUR_WORKFLOW_URL)
        >>> rag_agent.chat(messages=[{"role":"human", "content":"What is Clarifai"}])
    """
  chat_state_id = None

  def __init__(self,
               workflow_url: str = None,
               workflow: Workflow = None,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               **kwargs):
    """Initialize an empty or existing RAG.
    """
    self.logger = logger
    if workflow_url is not None and workflow is None:
      self.logger.info("workflow_url:%s", workflow_url)
      w = Workflow(workflow_url, base_url=base_url, pat=pat)
      self._prompt_workflow = w
      self._app = App(app_id=w.app_id, base_url=w.base, pat=w.pat)
    elif workflow_url is None and workflow is not None:
      self._prompt_workflow = workflow
      self._app = App(app_id=workflow.app_id, base_url=workflow.base, pat=workflow.pat)

  @classmethod
  def setup(cls,
            user_id: str = None,
            app_url: str = None,
            llm_url: str = "https://clarifai.com/mistralai/completion/models/mistral-7B-Instruct",
            base_workflow: str = "Text",
            workflow_yaml_filename: str = 'prompter_wf.yaml',
            workflow_id: str = None,
            base_url: str = "https://api.clarifai.com",
            pat: str = None,
            **kwargs):
    """Creates an app with `Text` as base workflow, create prompt model, create prompt workflow.

    **kwargs: Additional keyword arguments to be passed to rag-promter model.
          - min_score (float): The minimum score for search hits.
          - max_results (float): The maximum number of search hits.
          - prompt_template (str): The prompt template used. Must contain {data.hits} for the search hits and {data.text.raw} for the query string.

    Example:
        >>> from clarifai.rag import RAG
        >>> rag_agent = RAG.setup(user_id=YOUR_USER_ID)
        >>> rag_agent.chat(messages=[{"role":"human", "content":"What is Clarifai"}])

    Or if you already have an existing app with ingested data:
        >>> rag_agent = RAG.setup(app_url=YOUR_APP_URL)
        >>> rag_agent.chat(messages=[{"role":"human", "content":"What is Clarifai"}])
    """
    if not app_url:
      try:
        user_id = get_from_dict_or_env(key="user_id", env_key=CLARIFAI_USER_ID_ENV_VAR, **kwargs)
      except Exception:
        pass

    now_ts = uuid.uuid4().hex[:10]
    if user_id and not app_url:
      user = User(user_id=user_id, base_url=base_url, pat=pat)
      ## Create an App
      app_id = f"rag_app_{now_ts}"
      app = user.create_app(app_id=app_id, base_workflow=base_workflow)

    if not user_id and app_url:
      app = App(url=app_url, pat=pat)
      uid = app_url.split(".com/")[1].split("/")[0]
      user = User(user_id=uid, base_url=base_url, pat=pat)

    if user_id and app_url:
      raise UserError("Must provide one of user_id or app_url, not both.")

    if not user_id and not app_url:
      raise UserError(
          "user_id or app_url must be provided. The user_id can be found at https://clarifai.com/settings."
      )

    llm = Model(url=llm_url, pat=pat)

    min_score = kwargs.get("min_score", 0.95)
    max_results = kwargs.get("max_results", 5)
    prompt_template = kwargs.get("prompt_template", DEFAULT_RAG_PROMPT_TEMPLATE)
    params = Struct()
    params.update({
        "min_score": min_score,
        "max_results": max_results,
        "prompt_template": prompt_template
    })
    prompter_model_params = {"params": params}

    ## Create rag-prompter model and version
    model_id = f"prompter-{workflow_id}-{now_ts}" if workflow_id is not None else f"rag-prompter-{now_ts}"
    prompter_model = app.create_model(model_id=model_id, model_type_id="rag-prompter")
    prompter_model = prompter_model.create_version(output_info=prompter_model_params)

    ## Generate a tmp yaml file for workflow creation
    workflow_id = f"rag-wf-{now_ts}" if workflow_id is None else workflow_id
    workflow_dict = {
        "workflow": {
            "id":
                workflow_id,
            "nodes": [{
                "id": "rag-prompter",
                "model": {
                    "model_id": prompter_model.id,
                    "model_version_id": prompter_model.model_version.id
                }
            }, {
                "id": "llm",
                "model": {
                    "model_id": llm.id,
                    "user_id": llm.user_id,
                    "app_id": llm.app_id
                },
                "node_inputs": [{
                    "node_id": "rag-prompter"
                }]
            }]
        }
    }
    with open(workflow_yaml_filename, 'w') as out_file:
      yaml.dump(workflow_dict, out_file, default_flow_style=False)

    ## Create prompt workflow
    wf = app.create_workflow(config_filepath=workflow_yaml_filename)
    del user, llm, prompter_model, prompter_model_params
    return cls(workflow=wf)

  def upload(self,
             file_path: str = None,
             folder_path: str = None,
             url: str = None,
             batch_size: int = 128,
             chunk_size: int = 1024,
             chunk_overlap: int = 200,
             dataset_id: str = None,
             metadata: dict = None,
             **kwargs) -> None:
    """Uploads documents to the app.
        - Read from a local directory or public url or local filename.
        - Parse the document(s) into chunks.
        - Ingest chunks into the app with metadata.

    Args:
        file_path str: File path to the document.
        folder_path str: Folder path to the documents.
        url str: Public url to the document.
        batch_size int: Batch size for uploading.
        chunk_size int: Chunk size for splitting the document.
        chunk_overlap int: The token overlap of each chunk when splitting.
        **kwargs: Additional arguments for the SentenceSplitter. Refer https://docs.llamaindex.ai/en/stable/api/llama_index.node_parser.SentenceSplitter.html

    Example:
        >>> from clarifai.rag import RAG
        >>> rag_agent = RAG.setup(user_id=YOUR_USER_ID)
        >>> rag_agent.upload(folder_path = "~/work/docs")
        >>> rag_agent.upload(file_path = "~/work/docs/manual.pdf")
        >>> rag_agent.chat(messages=[{"role":"human", "content":"What is Clarifai"}])
    """
    #set batch size
    if batch_size > MAX_UPLOAD_BATCH_SIZE:
      raise ValueError(f"batch_size cannot be greater than {MAX_UPLOAD_BATCH_SIZE}")

    #check if only one of file_path, folder_path, or url is specified
    if file_path and (folder_path or url):
      raise ValueError("Only one of file_path, folder_path, or url can be specified.")
    if folder_path and (file_path or url):
      raise ValueError("Only one of file_path, folder_path, or url can be specified.")
    if url and (file_path or folder_path):
      raise ValueError("Only one of file_path, folder_path, or url can be specified.")

    #loading documents
    documents = load_documents(file_path=file_path, folder_path=folder_path, url=url)

    #splitting documents into chunks
    text_chunks = []
    metadata_list = []

    #iterate through documents
    for doc in documents:
      doc_i = 0
      cur_text_chunks = split_document(
          text=doc.text, chunk_size=chunk_size, chunk_overlap=chunk_overlap, **kwargs)
      text_chunks.extend(cur_text_chunks)
      metadata_list.extend([doc.metadata for _ in range(len(cur_text_chunks))])
      #if batch size is reached, upload the batch
      if len(text_chunks) > batch_size:
        for idx in range(0, len(text_chunks), batch_size):
          if idx + batch_size > len(text_chunks):
            continue
          batch_texts = text_chunks[0:batch_size]
          batch_ids = [uuid.uuid4().hex for _ in range(batch_size)]
          #metadata
          batch_metadatas = metadata_list[0:batch_size]
          meta_list = []
          for meta in batch_metadatas:
            meta_struct = Struct()
            meta_struct.update(meta)
            meta_struct.update({"doc_chunk_no": doc_i})
            if metadata and isinstance(metadata, dict):
              meta_struct.update(metadata)
            meta_list.append(meta_struct)
            doc_i += 1
          del batch_metadatas
          #creating input proto
          input_batch = [
              self._app.inputs().get_text_input(
                  input_id=batch_ids[i],
                  raw_text=text,
                  dataset_id=dataset_id,
                  metadata=meta_list[i],
              ) for i, text in enumerate(batch_texts)
          ]
          #uploading input with metadata
          self._app.inputs().upload_inputs(inputs=input_batch)
          #delete uploaded chunks
          del text_chunks[0:batch_size]
          del metadata_list[0:batch_size]

    #uploading the remaining chunks
    if len(text_chunks) > 0:
      batch_size = len(text_chunks)
      batch_ids = [uuid.uuid4().hex for _ in range(batch_size)]
      #metadata
      batch_metadatas = metadata_list[0:batch_size]
      meta_list = []
      for meta in batch_metadatas:
        meta_struct = Struct()
        meta_struct.update(meta)
        meta_struct.update({"doc_chunk_no": doc_i})
        if metadata and isinstance(metadata, dict):
          meta_struct.update(metadata)
        meta_list.append(meta_struct)
        doc_i += 1
      del batch_metadatas
      #creating input proto
      input_batch = [
          self._app.inputs().get_text_input(
              input_id=batch_ids[i],
              raw_text=text,
              dataset_id=dataset_id,
              metadata=meta_list[i],
          ) for i, text in enumerate(text_chunks)
      ]
      #uploading input with metadata
      self._app.inputs().upload_inputs(inputs=input_batch)
      del text_chunks
      del metadata_list

  def chat(self, messages: List[dict], client_manage_state: bool = False) -> List[dict]:
    """Chat interface in OpenAI API format.

    Args:
        messages List[dict]: A list of dictionary in the following format:
        ```
        [
          {"role": "user", "content": "Hello there."},
          {"role": "assistant", "content": "Hi, I'm Claude. How can I help you?"},
          {"role": "user", "content": "Can you explain LLMs in plain English?"},
        ]
        ```
        client_manage_state (bool): Whether the client will handle chat state management. Default is false.

    This will pass back the workflow state ID for the server to store chat state.
    """
    if client_manage_state:
      single_prompt = convert_messages_to_str(messages)
      input_proto = Inputs._get_proto("", "", text_pb=resources_pb2.Text(raw=single_prompt))
      response = self._prompt_workflow.predict([input_proto])
      messages.append(format_assistant_message(response.results[0].outputs[-1].data.text.raw))
      return messages

    # server-side state management
    message = messages[-1].get("content", "")
    if len(message) == 0:
      raise UserError("Empty message supplied.")

    # get chat state id
    chat_state_id = "init" if self.chat_state_id is None else self.chat_state_id

    # call predict
    input_proto = Inputs._get_proto("", "", text_pb=resources_pb2.Text(raw=message))
    response = self._prompt_workflow.predict([input_proto], workflow_state_id=chat_state_id)

    # store chat state id
    self.chat_state_id = response.workflow_state.id
    return [format_assistant_message(response.results[0].outputs[-1].data.text.raw)]



#! ImageNet Classification dataset

import os

from clarifai.datasets.upload.base import ClarifaiDataLoader
from ..features import VisualClassificationFeatures


class ImageNetDataLoader(ClarifaiDataLoader):
  """ImageNet Dataset."""

  def __init__(self, data_dir, split: str = "train"):
    """
    Initialize dataset params.
    Args:
      data_dir: the local dataset directory.
      split: "train" or "test"
    """
    self.split = split
    self.data_dir = data_dir
    self.label_map = dict()
    self.concepts = []
    self.image_paths = []

    self.load_data()

  def load_data(self):
    #Creating label map
    with open(os.path.join(self.data_dir, "LOC_synset_mapping.txt")) as _file:
      for _id in _file:
        #Removing the spaces,upper quotes and Converting to set to remove repetitions. Then converting to list for compatibility.
        self.label_map[_id.split(" ")[0]] = list({
            "".join(("".join((label.rstrip().lstrip().split(" ")))).split("'"))
            for label in _id[_id.find(" ") + 1:].split(",")
        })

    for _folder in os.listdir(os.path.join(self.data_dir, self.split)):
      try:
        concept = self.label_map[_folder]  #concepts
      except Exception:
        continue
      folder_path = os.path.join(self.data_dir, self.split) + "/" + _folder
      for _img in os.listdir(folder_path):
        if _img.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff')):
          self.concepts.append(concept)
          self.image_paths.append(folder_path + "/" + _img)

    assert len(self.concepts) == len(self.image_paths)
    "Number of concepts and images are not equal"

  def __len__(self):
    return len(self.image_paths)

  def __getitem__(self, idx):
    return VisualClassificationFeatures(
        image_path=self.image_paths[idx],
        label=self.concepts[idx],
        id=self.image_paths[idx].split('.')[0].split('/')[-1])



from typing import Any, Dict

import yaml
from google.protobuf.json_format import MessageToDict

VALID_YAML_KEYS = ["workflow", "id", "nodes", "node_inputs", "node_id", "model"]


def clean_up_unused_keys(wf: dict):
  """Removes unused keys from dict before exporting to yaml. Supports nested dicts."""
  new_wf = dict()
  for key, val in wf.items():
    if key not in VALID_YAML_KEYS:
      continue
    if key == "model":
      new_wf["model"] = {
          "model_id": wf["model"]["id"],
          "model_version_id": wf["model"]["model_version"]["id"]
      }
      # If the model is not from clarifai main, add the app_id and user_id to the model dict.
      if wf["model"]["user_id"] != "clarifai" and wf["model"]["app_id"] != "main":
        new_wf["model"].update({
            "app_id": wf["model"]["app_id"],
            "user_id": wf["model"]["user_id"]
        })
    elif isinstance(val, dict):
      new_wf[key] = clean_up_unused_keys(val)
    elif isinstance(val, list):
      new_list = []
      for i in val:
        new_list.append(clean_up_unused_keys(i))
      new_wf[key] = new_list
    else:
      new_wf[key] = val
  return new_wf


class Exporter:

  def __init__(self, workflow):
    self.wf = workflow

  def __enter__(self):
    return self

  def parse(self) -> Dict[str, Any]:
    """Reads a resources_pb2.Workflow object (e.g. from a GetWorkflow response)

    Returns:
        dict: A dict representation of the workflow.
    """
    if isinstance(self.wf, list):
      self.wf = self.wf[0]
    wf = {"workflow": MessageToDict(self.wf, preserving_proto_field_name=True)}
    clean_wf = clean_up_unused_keys(wf)
    self.wf_dict = clean_wf
    return clean_wf

  def export(self, out_path):
    with open(out_path, 'w') as out_file:
      yaml.dump(self.wf_dict["workflow"], out_file, default_flow_style=False)

  def __exit__(self, *args):
    self.close()

  def close(self):
    del self.wf
    del self.wf_dict
